from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Sequence

from dbt_semantic_interfaces.references import SemanticModelReference


class SemanticModelDerivation(ABC):
    """Interface for an object that can be described as derived from a semantic model."""

    # There are some cases where we create virtual elements that don't directly correspond to something in the manifest.
    # For example, when querying `metric_time` without any metrics. To avoid having an `Optional` field to handle those
    # cases, we're trying out a case where we use this reference.
    VIRTUAL_SEMANTIC_MODEL_REFERENCE = SemanticModelReference("__VIRTUAL__")

    @property
    @abstractmethod
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        """The semantic models that this was derived from.

        The returned sequence should be ordered and not contain duplicates.
        """
        raise NotImplementedError



from __future__ import annotations

import dataclasses
import logging
import threading
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Dict, Generator

from typing_extensions import override

from metricflow_semantics.dag.id_prefix import IdPrefix

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SequentialId:
    """Returns a sequentially numbered ID based on a prefix."""

    id_prefix: IdPrefix
    index: int

    @property
    def str_value(self) -> str:  # noqa: D102
        return f"{self.id_prefix.str_value}_{self.index}"

    @override
    def __repr__(self) -> str:
        return self.str_value


@dataclass
class _IdGenerationState:
    """A thread-local class that keeps track of the next IDs to return for a given prefix."""

    default_start_value: int
    prefix_to_next_value: Dict[IdPrefix, int] = dataclasses.field(default_factory=dict)


class _IdGenerationStateStack(threading.local):
    """A thread-local stack that keeps track of the state of ID generation.

    The stack allows the use of context managers to enter sections where ID generation starts at a configured value.
    When entering the section, a new `_IdGenerationState` is pushed on to the stack, and the state at the top of the
    stack is used to generate IDs. When exiting a section, the state is popped off so that ID generation resumes from
    the previous values.

    This stack is thread-local so that ID generation is consistent for a given thread.
    """

    def __init__(self, initial_default_start_value: int = 0) -> None:  # noqa: D107
        self._state_stack = [_IdGenerationState(initial_default_start_value)]

    def push_state(self, id_generation_state: _IdGenerationState) -> None:
        self._state_stack.append(id_generation_state)

    def pop_state(self) -> None:
        state_stack_size = len(self._state_stack)
        if state_stack_size <= 1:
            logger.error(
                f"Attempted to pop the stack when {state_stack_size=}. Since sequential ID generation may not "
                f"be absolutely critical for resolving queries, logging this as an error but it should be "
                f"investigated.",
                stack_info=True,
            )
            return

        self._state_stack.pop(-1)

    @property
    def current_state(self) -> _IdGenerationState:
        return self._state_stack[-1]


class SequentialIdGenerator:
    """Generates sequential ID values based on a prefix."""

    _THREAD_LOCAL_ID_GENERATION_STATE_STACK = _IdGenerationStateStack()

    @classmethod
    def create_next_id(cls, id_prefix: IdPrefix) -> SequentialId:  # noqa: D102
        id_generation_state = cls._THREAD_LOCAL_ID_GENERATION_STATE_STACK.current_state
        if id_prefix not in id_generation_state.prefix_to_next_value:
            id_generation_state.prefix_to_next_value[id_prefix] = id_generation_state.default_start_value
        index = id_generation_state.prefix_to_next_value[id_prefix]
        id_generation_state.prefix_to_next_value[id_prefix] = index + 1
        return SequentialId(id_prefix, index)

    @classmethod
    def reset(cls, default_start_value: int = 0) -> None:
        """Resets the numbering of the generated IDs so that it starts at the given value."""
        id_generation_state = cls._THREAD_LOCAL_ID_GENERATION_STATE_STACK.current_state
        id_generation_state.prefix_to_next_value = {}
        id_generation_state.default_start_value = default_start_value

    @classmethod
    @contextmanager
    def id_number_space(cls, start_value: int) -> Generator[None, None, None]:
        """Open a context where ID generation starts with the given start value.

        On exit, resume ID numbering from prior to entering the context.
        """
        SequentialIdGenerator._THREAD_LOCAL_ID_GENERATION_STATE_STACK.push_state(_IdGenerationState(start_value))
        yield None
        SequentialIdGenerator._THREAD_LOCAL_ID_GENERATION_STATE_STACK.pop_state()



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import ResolverInputForQuery


@dataclass(frozen=True)
class NoMetricOrGroupByIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue with the query where there are no metrics or group by inputs in the query."""

    resolver_input_for_query: ResolverInputForQuery

    @staticmethod
    def from_parameters(  # noqa: D102
        resolver_input_for_query: ResolverInputForQuery, query_resolution_path: MetricFlowQueryResolutionPath
    ) -> NoMetricOrGroupByIssue:
        return NoMetricOrGroupByIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            resolver_input_for_query=resolver_input_for_query,
            query_resolution_path=query_resolution_path,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return "There are no metrics or group by items requested in the query."

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> NoMetricOrGroupByIssue:
        return NoMetricOrGroupByIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            resolver_input_for_query=self.resolver_input_for_query,
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
        )



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import Iterator, Sequence, Tuple, Type

from metricflow.data_table.column_types import CellValue


@dataclass(frozen=True)
class ColumnDescription:
    """Describes a single column in a data table."""

    column_name: str
    column_type: Type[CellValue]

    def with_lower_case_column_name(self) -> ColumnDescription:  # noqa: D102
        return ColumnDescription(
            column_name=self.column_name.lower(),
            column_type=self.column_type,
        )


@dataclass(frozen=True)
class ColumnDescriptionSet:
    """Describes a collection of columns in a data table."""

    column_descriptions: Tuple[ColumnDescription, ...]

    def __iter__(self) -> Iterator[ColumnDescription]:  # noqa: D105
        return iter(self.column_descriptions)

    @cached_property
    def column_names(self) -> Sequence[str]:  # noqa: D102
        return tuple(column_description.column_name for column_description in self.column_descriptions)

    @cached_property
    def column_types(self) -> Sequence[Type]:  # noqa: D102
        return tuple(column_description.column_type for column_description in self.column_descriptions)



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import override

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import (
    GroupByItemResolutionNode,
    GroupByItemResolutionNodeSet,
    GroupByItemResolutionNodeVisitor,
)
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class NoMetricsGroupByItemSourceNode(GroupByItemResolutionNode):
    """Outputs group-by-items that can be queried without any metrics."""

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 0

    @staticmethod
    def create() -> NoMetricsGroupByItemSourceNode:  # noqa: D102
        return NoMetricsGroupByItemSourceNode(parent_nodes=())

    @override
    def accept(self, visitor: GroupByItemResolutionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_no_metrics_query_node(self)

    @property
    @override
    def description(self) -> str:
        return "Output the available group-by-items for a query without any metrics."

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.VALUES_GROUP_BY_ITEM_RESOLUTION_NODE

    @property
    @override
    def ui_description(self) -> str:
        return f"{self.__class__.__name__}()"

    @override
    def _self_set(self) -> GroupByItemResolutionNodeSet:
        return GroupByItemResolutionNodeSet(no_metrics_query_nodes=(self,))



from __future__ import annotations

import textwrap
from typing import Dict, Optional


class CustomerFacingSemanticException(Exception):
    """Class of Exceptions that make it to the customer's eyeballs."""

    pass


class UnableToSatisfyQueryError(CustomerFacingSemanticException):  # noqa: D101
    def __init__(self, error_name: str, context: Optional[Dict[str, str]] = None) -> None:
        """Context will be printed as list of items when this is converted to a string."""
        self.error_name = error_name
        self._context = context

    def __str__(self) -> str:  # noqa: D105
        error_lines = ["Unable To Satisfy Query Error: " + self.error_name]
        if self._context:
            for key, value in self._context.items():
                error_lines.append(f"\n{key}:")
                error_lines.append(f"{textwrap.indent(value, prefix='    ')}")
        return "\n".join(error_lines)


class SemanticException(Exception):  # noqa: D101
    pass


class DuplicateMetricError(SemanticException):  # noqa: D101
    pass


class MetricNotFoundError(SemanticException, KeyError):  # noqa: D101
    pass


class NonExistentMeasureError(SemanticException):  # noqa: D101
    pass


class InvalidSemanticModelError(SemanticException):  # noqa: D101
    pass


class ExecutionException(Exception):
    """Raised if there are any errors while executing the execution plan."""

    pass


class ModelCreationException(Exception):
    """Exception to represent errors related to the building a model."""

    def __init__(self, msg: str = "") -> None:  # noqa: D107
        error_msg = "An error occurred when attempting to build the semantic model"
        if msg:
            error_msg += f"\n{msg}"
        super().__init__(error_msg)


class InferenceError(Exception):
    """Exception to represent errors related to inference."""


class UnsupportedEngineFeatureError(RuntimeError):
    """Raised when the user attempts to use a feature that isn't supported by the data platform."""


class SqlBindParametersNotSupportedError(Exception):
    """Raised when a SqlClient that does not have support for bind parameters receives a non-empty set of params."""


class UnknownMetricLinkingError(ValueError):
    """Raised during linking when a user attempts to use a metric that isn't specified."""


class InvalidQuerySyntax(Exception):
    """Raised when query syntax is invalid. Primarily used in the where clause."""

    def __init__(self, msg: str) -> None:  # noqa: D107
        super().__init__(msg)



from __future__ import annotations

import datetime
import decimal
from typing import Sequence, Tuple, Type, Union

# Types supported by `MetricFlowDataTable`.
CellValue = Union[float, str, datetime.datetime, bool, None]
# Types supported as inputs when building a `MetricFlowDataTable`. These inputs will get converted into
# one of the `CellValue` types.
InputCellValue = Union[int, float, str, datetime.datetime, bool, None, decimal.Decimal, datetime.date]
MetricflowNoneType = type(None)


def row_cell_types(row: Sequence[CellValue]) -> Tuple[Type[CellValue], ...]:
    """Return the cell type / column type for the objects in the row."""
    return tuple(cell_type(cell) for cell in row)


def cell_type(cell_value: CellValue) -> Type[CellValue]:
    """Return the cell type / column type for the object in the cell."""
    return type(cell_value)



from __future__ import annotations

from typing import Sequence

from typing_extensions import override

from metricflow_semantics.query.group_by_item.candidate_push_down.push_down_visitor import DagTraversalPathTracker
from metricflow_semantics.query.group_by_item.resolution_dag.dag import GroupByItemResolutionDag
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import (
    GroupByItemResolutionNode,
    GroupByItemResolutionNodeVisitor,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.measure_source_node import (
    MeasureGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.metric_resolution_node import (
    MetricGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.no_metrics_query_source_node import (
    NoMetricsGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
    QueryGroupByItemResolutionNode,
)
from metricflow_semantics.query.issues.issues_base import MetricFlowQueryResolutionIssueSet
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import ResolverInputForQuery
from metricflow_semantics.query.validation_rules.base_validation_rule import PostResolutionQueryValidationRule


class PostResolutionQueryValidator:
    """Runs query validation rules after query resolution is complete."""

    def validate_query(
        self,
        resolution_dag: GroupByItemResolutionDag,
        resolver_input_for_query: ResolverInputForQuery,
        validation_rules: Sequence[PostResolutionQueryValidationRule],
    ) -> MetricFlowQueryResolutionIssueSet:
        """Validate according to the list of configured validation rules and return a set containing issues found."""
        validation_visitor = _PostResolutionQueryValidationVisitor(
            resolver_input_for_query=resolver_input_for_query,
            validation_rules=validation_rules,
        )

        return resolution_dag.sink_node.accept(validation_visitor)


class _PostResolutionQueryValidationVisitor(GroupByItemResolutionNodeVisitor[MetricFlowQueryResolutionIssueSet]):
    """Visitor that runs the validation rule when it visits a metric."""

    def __init__(
        self,
        resolver_input_for_query: ResolverInputForQuery,
        validation_rules: Sequence[PostResolutionQueryValidationRule],
    ) -> None:
        self._validation_rules = validation_rules
        self._path_from_start_node_tracker = DagTraversalPathTracker()
        self._resolver_input_for_query = resolver_input_for_query

    def _default_handler(self, node: GroupByItemResolutionNode) -> MetricFlowQueryResolutionIssueSet:
        with self._path_from_start_node_tracker.track_node_visit(node):
            return MetricFlowQueryResolutionIssueSet.merge_iterable(
                parent_node.accept(self) for parent_node in node.parent_nodes
            )

    @override
    def visit_measure_node(self, node: MeasureGroupByItemSourceNode) -> MetricFlowQueryResolutionIssueSet:
        with self._path_from_start_node_tracker.track_node_visit(node) as current_traversal_path:
            issue_sets_to_merge = [parent_node.accept(self) for parent_node in node.parent_nodes]

            for validation_rule in self._validation_rules:
                issue_sets_to_merge.append(
                    validation_rule.validate_measure_in_resolution_dag(
                        measure_reference=node.measure_reference,
                        resolution_path=current_traversal_path,
                    )
                )

            return MetricFlowQueryResolutionIssueSet.merge_iterable(issue_sets_to_merge)

    @override
    def visit_metric_node(self, node: MetricGroupByItemResolutionNode) -> MetricFlowQueryResolutionIssueSet:
        with self._path_from_start_node_tracker.track_node_visit(node) as current_traversal_path:
            issue_sets_to_merge = [parent_node.accept(self) for parent_node in node.parent_nodes]

            for validation_rule in self._validation_rules:
                issue_sets_to_merge.append(
                    validation_rule.validate_metric_in_resolution_dag(
                        metric_reference=node.metric_reference,
                        resolution_path=current_traversal_path,
                    )
                )

            return MetricFlowQueryResolutionIssueSet.merge_iterable(issue_sets_to_merge)

    @override
    def visit_query_node(self, node: QueryGroupByItemResolutionNode) -> MetricFlowQueryResolutionIssueSet:
        with self._path_from_start_node_tracker.track_node_visit(node) as current_traversal_path:
            issue_sets_to_merge = [parent_node.accept(self) for parent_node in node.parent_nodes]

            for validation_rule in self._validation_rules:
                issue_sets_to_merge.append(
                    validation_rule.validate_query_in_resolution_dag(
                        metrics_in_query=node.metrics_in_query,
                        where_filter_intersection=node.where_filter_intersection,
                        resolution_path=current_traversal_path,
                    )
                )

            return MetricFlowQueryResolutionIssueSet.merge_iterable(issue_sets_to_merge)

    @override
    def visit_no_metrics_query_node(self, node: NoMetricsGroupByItemSourceNode) -> MetricFlowQueryResolutionIssueSet:
        return self._default_handler(node)



from __future__ import annotations

import textwrap


def mf_dedent(text: str) -> str:
    """Remove leading newlines, dedents, and remove tailing newlines.

    This function simplifies the somewhat-frequently used:

        text = textwrap.dedent(
            [triple quote][backslash]
            Line 0
            Line 1
            [triple quote]
        ).rstrip()

    to:

       text = mf_dedent(
           [triple quote]
           Line 0
           Line 1
           [triple quote]
       )
    """
    return textwrap.dedent(text.lstrip("\n")).rstrip("\n")



from __future__ import annotations

from abc import ABC, abstractmethod

from dbt_semantic_interfaces.call_parameter_sets import (
    DimensionCallParameterSet,
    EntityCallParameterSet,
    MetricCallParameterSet,
    TimeDimensionCallParameterSet,
)
from typing_extensions import override

from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.patterns.typed_patterns import (
    DimensionPattern,
    EntityPattern,
    GroupByMetricPattern,
    TimeDimensionPattern,
)


class WhereFilterPatternFactory(ABC):
    """Interface that defines how spec patterns should be generated for the group-by-items specified in filters."""

    @abstractmethod
    def create_for_dimension_call_parameter_set(  # noqa: D102
        self, dimension_call_parameter_set: DimensionCallParameterSet
    ) -> SpecPattern:
        raise NotImplementedError

    @abstractmethod
    def create_for_time_dimension_call_parameter_set(  # noqa: D102
        self, time_dimension_call_parameter_set: TimeDimensionCallParameterSet
    ) -> SpecPattern:
        raise NotImplementedError

    @abstractmethod
    def create_for_entity_call_parameter_set(  # noqa: D102
        self, entity_call_parameter_set: EntityCallParameterSet
    ) -> SpecPattern:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def create_for_metric_call_parameter_set(  # noqa: D102
        self, metric_call_parameter_set: MetricCallParameterSet
    ) -> SpecPattern:
        raise NotImplementedError


class DefaultWhereFilterPatternFactory(WhereFilterPatternFactory):
    """Default implementation using patterns derived from EntityLinkPattern."""

    @override
    def create_for_dimension_call_parameter_set(
        self, dimension_call_parameter_set: DimensionCallParameterSet
    ) -> SpecPattern:
        return DimensionPattern.from_call_parameter_set(dimension_call_parameter_set)

    @override
    def create_for_time_dimension_call_parameter_set(
        self, time_dimension_call_parameter_set: TimeDimensionCallParameterSet
    ) -> SpecPattern:
        return TimeDimensionPattern.from_call_parameter_set(time_dimension_call_parameter_set)

    @override
    def create_for_entity_call_parameter_set(self, entity_call_parameter_set: EntityCallParameterSet) -> SpecPattern:
        return EntityPattern.from_call_parameter_set(entity_call_parameter_set)

    @override
    def create_for_metric_call_parameter_set(self, metric_call_parameter_set: MetricCallParameterSet) -> SpecPattern:
        return GroupByMetricPattern.from_call_parameter_set(metric_call_parameter_set)



from __future__ import annotations

from typing import TYPE_CHECKING, List

from dbt_semantic_interfaces.references import EntityReference, SemanticModelElementReference, SemanticModelReference
from metricflow_semantics.instances import EntityInstance, InstanceSet
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.model.semantics.semantic_model_join_evaluator import SemanticModelJoinEvaluator

if TYPE_CHECKING:
    from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup


class JoinDataflowOutputValidator:
    """Checks that the instances in the output of a join dataflow node is valid."""

    def __init__(self, semantic_model_lookup: SemanticModelLookup) -> None:  # noqa: D107
        self._join_evaluator = SemanticModelJoinEvaluator(semantic_model_lookup)

    @staticmethod
    def _semantic_model_of_entity_in_instance_set(
        instance_set: InstanceSet,
        entity_reference: EntityReference,
    ) -> SemanticModelReference:
        """Return the semantic model where the entity was defined in the instance set."""
        matching_instances: List[EntityInstance] = []
        for entity_instance in instance_set.entity_instances:
            assert len(entity_instance.defined_from) == 1
            if len(entity_instance.spec.entity_links) == 0 and entity_instance.spec.reference == entity_reference:
                matching_instances.append(entity_instance)

        assert len(matching_instances) == 1, (
            f"Not exactly 1 matching entity instances found: {matching_instances} for {entity_reference} in "
            f"{mf_pformat(instance_set)}"
        )
        return matching_instances[0].origin_semantic_model_reference.semantic_model_reference

    def is_valid_instance_set_join(
        self,
        left_instance_set: InstanceSet,
        right_instance_set: InstanceSet,
        on_entity_reference: EntityReference,
        right_node_is_aggregated_to_entity: bool = False,
    ) -> bool:
        """Return true if the instance sets can be joined using the given entity."""
        left_semantic_model_reference = self._semantic_model_of_entity_in_instance_set(
            instance_set=left_instance_set, entity_reference=on_entity_reference
        )
        if right_node_is_aggregated_to_entity:
            left_entity = self._join_evaluator._semantic_model_lookup.get_entity_in_semantic_model(
                SemanticModelElementReference.create_from_references(left_semantic_model_reference, on_entity_reference)
            )
            if not left_entity:
                return False
            possible_right_entities = [
                entity_instance
                for entity_instance in right_instance_set.entity_instances
                if entity_instance.spec.reference == on_entity_reference
            ]
            if len(possible_right_entities) != 1:
                return False

            # No fan-out check needed since right subquery is aggregated to the entity level, ensuring uniqueness.
            return True
        else:
            return self._join_evaluator.is_valid_semantic_model_join(
                left_semantic_model_reference=JoinDataflowOutputValidator._semantic_model_of_entity_in_instance_set(
                    instance_set=left_instance_set, entity_reference=on_entity_reference
                ),
                right_semantic_model_reference=JoinDataflowOutputValidator._semantic_model_of_entity_in_instance_set(
                    instance_set=right_instance_set,
                    entity_reference=on_entity_reference,
                ),
                on_entity_reference=on_entity_reference,
            )



from __future__ import annotations

from metricflow_semantics.specs.spec_set import InstanceSpecSet, InstanceSpecSetTransform


class MatchingLinkableSpecsTransform(InstanceSpecSetTransform[bool]):
    """Returns true if two spec sets have the same set of linkable specs."""

    def __init__(self, left_spec_set: InstanceSpecSet) -> None:  # noqa: D107
        self._left_spec_set = left_spec_set

    def transform(self, spec_set: InstanceSpecSet) -> bool:  # noqa: D102
        return (
            set(self._left_spec_set.dimension_specs) == set(spec_set.dimension_specs)
            and set(self._left_spec_set.time_dimension_specs) == set(spec_set.time_dimension_specs)
            and set(self._left_spec_set.entity_specs) == set(spec_set.entity_specs)
            and set(self._left_spec_set.group_by_metric_specs) == set(spec_set.group_by_metric_specs)
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from dbt_semantic_interfaces.naming.keywords import METRIC_TIME_ELEMENT_NAME
from typing_extensions import override

from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.spec_set import group_specs_by_type
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec


@dataclass(frozen=True)
class MetricTimePattern(SpecPattern):
    """Pattern that matches to only metric_time specs.

    This pattern can be used to check if metric_time has been specified in a query, or to help implement checks that
    only apply to metric_time specs.
    """

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[TimeDimensionSpec]:
        spec_set = group_specs_by_type(candidate_specs)
        return tuple(
            time_dimension_spec
            for time_dimension_spec in spec_set.time_dimension_specs
            if time_dimension_spec.element_name == METRIC_TIME_ELEMENT_NAME
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Sequence

from typing_extensions import override

from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.spec_set import group_specs_by_type


@dataclass(frozen=True)
class NoneDatePartPattern(SpecPattern):
    """Matches to linkable specs, but for time dimension specs, only matches to ones without date_part.

    This is used to help implement restrictions for cumulative metrics where those can not be queried by date_part.
    """

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[LinkableInstanceSpec]:
        specs_to_return: List[LinkableInstanceSpec] = []
        spec_set = group_specs_by_type(candidate_specs)
        for time_dimension_spec in spec_set.time_dimension_specs:
            if time_dimension_spec.date_part is None:
                specs_to_return.append(time_dimension_spec)
        specs_to_return.extend(spec_set.dimension_specs)
        specs_to_return.extend(spec_set.entity_specs)
        specs_to_return.extend(spec_set.group_by_metric_specs)

        return specs_to_return



from __future__ import annotations

import logging
from typing import Optional, Sequence, Tuple

from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilterIntersection
from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from dbt_semantic_interfaces.references import MeasureReference, MetricReference
from dbt_semantic_interfaces.type_enums import MetricType

from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.query.group_by_item.resolution_dag.dag import GroupByItemResolutionDag
from metricflow_semantics.query.group_by_item.resolution_dag.input_metric_location import InputMetricDefinitionLocation
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.measure_source_node import (
    MeasureGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.metric_resolution_node import (
    MetricGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.no_metrics_query_source_node import (
    NoMetricsGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
    QueryGroupByItemResolutionNode,
)

logger = logging.getLogger(__name__)


class GroupByItemResolutionDagBuilder:
    """Builds a GroupByItemResolutionDag that can be used to resolve group-by-item specs from spec patterns."""

    def __init__(self, manifest_lookup: SemanticManifestLookup) -> None:  # noqa: D107
        self._manifest_lookup = manifest_lookup

    def _build_dag_component_for_metric(
        self,
        metric_reference: MetricReference,
        metric_input_location: Optional[InputMetricDefinitionLocation],
    ) -> MetricGroupByItemResolutionNode:
        """Builds a DAG component that represents the resolution flow for a metric."""
        metric = self._manifest_lookup.metric_lookup.get_metric(metric_reference)

        # For a base metric, the parents are measure nodes
        if len(metric.input_metrics) == 0:
            measure_references_for_metric: Tuple[MeasureReference, ...]
            if metric.type is MetricType.CONVERSION:
                conversion_type_params = metric.type_params.conversion_type_params
                assert (
                    conversion_type_params
                ), "A conversion metric should have type_params.conversion_type_params defined."
                measure_references_for_metric = (conversion_type_params.base_measure.measure_reference,)
            else:
                measure_references_for_metric = tuple(
                    input_measure.measure_reference for input_measure in metric.input_measures
                )

            source_candidates_for_measure_nodes = tuple(
                MeasureGroupByItemSourceNode.create(
                    measure_reference=measure_reference,
                    child_metric_reference=metric_reference,
                )
                for measure_reference in measure_references_for_metric
            )
            return MetricGroupByItemResolutionNode.create(
                metric_reference=metric_reference,
                metric_input_location=metric_input_location,
                parent_nodes=source_candidates_for_measure_nodes,
            )
        # For a derived metric, the parents are other metrics.
        return MetricGroupByItemResolutionNode.create(
            metric_reference=metric_reference,
            metric_input_location=metric_input_location,
            parent_nodes=tuple(
                self._build_dag_component_for_metric(
                    metric_reference=metric_input.as_reference,
                    metric_input_location=InputMetricDefinitionLocation(
                        derived_metric_reference=metric_reference,
                        input_metric_list_index=metric_input_index,
                    ),
                )
                for metric_input_index, metric_input in enumerate(metric.input_metrics)
            ),
        )

    def _build_dag_component_for_query(
        self, metric_references: Sequence[MetricReference], where_filter_intersection: WhereFilterIntersection
    ) -> QueryGroupByItemResolutionNode:
        """Builds a DAG component that represents the resolution flow for a query."""
        if len(metric_references) == 0:
            return QueryGroupByItemResolutionNode.create(
                parent_nodes=(NoMetricsGroupByItemSourceNode.create(),),
                metrics_in_query=metric_references,
                where_filter_intersection=where_filter_intersection,
            )
        return QueryGroupByItemResolutionNode.create(
            parent_nodes=tuple(
                self._build_dag_component_for_metric(
                    metric_reference=metric_reference,
                    metric_input_location=None,
                )
                for metric_reference in metric_references
            ),
            metrics_in_query=metric_references,
            where_filter_intersection=where_filter_intersection,
        )

    def build(
        self, metric_references: Sequence[MetricReference], where_filter_intersection: Optional[WhereFilterIntersection]
    ) -> GroupByItemResolutionDag:
        """Build a resolution DAG for a query.

        Args:
            metric_references: The metrics in the query.
            where_filter_intersection: The filters in the query.

        Returns:
            The associated group-by-item resolution DAG.
        """
        return GroupByItemResolutionDag(
            sink_node=self._build_dag_component_for_query(
                metric_references=metric_references,
                where_filter_intersection=where_filter_intersection
                or PydanticWhereFilterIntersection(where_filters=[]),
            )
        )



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.sql.sql_exprs import (
    SqlColumnAliasReferenceExpression,
    SqlColumnReference,
    SqlColumnReplacements,
    SqlExpressionNode,
    SqlExpressionTreeLineage,
    SqlLogicalExpression,
    SqlLogicalOperator,
)
from typing_extensions import override

from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlPlanOptimizer
from metricflow.sql.sql_plan import (
    SqlCreateTableAsNode,
    SqlCteNode,
    SqlJoinDescription,
    SqlOrderByDescription,
    SqlPlanNode,
    SqlPlanNodeVisitor,
    SqlSelectColumn,
    SqlSelectQueryFromClauseNode,
    SqlSelectStatementNode,
    SqlTableNode,
)

logger = logging.getLogger(__name__)


@dataclass
class RewritableSqlClauses:
    """Stores clauses in a SQL query that should be rewritten when a node is reduced."""

    select_columns: List[SqlSelectColumn]
    wheres: List[SqlExpressionNode]
    group_bys: List[SqlSelectColumn]
    order_bys: List[SqlOrderByDescription]

    def rewrite(self, column_replacements: SqlColumnReplacements) -> None:
        """Rewrite all clauses using the given replacements."""
        self.select_columns = [
            SqlSelectColumn(expr=x.expr.rewrite(column_replacements), column_alias=x.column_alias)
            for x in self.select_columns
        ]
        self.wheres = [x.rewrite(column_replacements=column_replacements) for x in self.wheres]
        self.group_bys = [
            SqlSelectColumn(expr=x.expr.rewrite(column_replacements), column_alias=x.column_alias)
            for x in self.group_bys
        ]
        self.order_bys = [
            SqlOrderByDescription(
                expr=x.expr.rewrite(column_replacements=column_replacements),
                desc=x.desc,
            )
            for x in self.order_bys
        ]

    def combine_wheres(self, additional_where_clauses: List[SqlExpressionNode]) -> Optional[SqlExpressionNode]:
        """Combine the WHERE clauses in this with the additional clauses to form a single WHERE clause."""
        all_where_clauses = self.wheres + additional_where_clauses
        if len(all_where_clauses) == 1:
            return all_where_clauses[0]
        elif len(all_where_clauses) > 1:
            return SqlLogicalExpression.create(
                operator=SqlLogicalOperator.AND,
                args=tuple(all_where_clauses),
            )
        return None

    @property
    def contains_ambiguous_exprs(self) -> bool:
        """Returns true if any of the clauses have ambiguous expressions that will be difficult to re-write."""
        return any(
            [x.expr.lineage.contains_ambiguous_exprs for x in self.select_columns]
            + [x.lineage.contains_ambiguous_exprs for x in self.wheres]
            + [x.expr.lineage.contains_ambiguous_exprs for x in self.group_bys]
            + [x.expr.lineage.contains_ambiguous_exprs for x in self.order_bys]
        )


class SqlRewritingSubQueryReducerVisitor(SqlPlanNodeVisitor[SqlPlanNode]):
    """Visits the SQL query plan to simplify sub-queries. On each visit, return a simplified node.

    Unlike SqlSubQueryReducerVisitor, this will re-write expressions to realize more reductions.
    """

    def _reduce_parents(
        self,
        node: SqlSelectStatementNode,
    ) -> SqlSelectStatementNode:
        """Apply the reducing operation to the parent select statements."""
        return SqlSelectStatementNode.create(
            description=node.description,
            select_columns=node.select_columns,
            from_source=node.from_source.accept(self),
            from_source_alias=node.from_source_alias,
            cte_sources=tuple(
                cte_source.with_new_select(cte_source.select_statement.accept(self)) for cte_source in node.cte_sources
            ),
            join_descs=tuple(
                SqlJoinDescription(
                    right_source=join_desc.right_source.accept(self),
                    right_source_alias=join_desc.right_source_alias,
                    on_condition=join_desc.on_condition,
                    join_type=join_desc.join_type,
                )
                for join_desc in node.join_descs
            ),
            group_bys=node.group_bys,
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
            distinct=node.distinct,
        )

    @staticmethod
    def _statement_contains_difficult_expressions(node: SqlSelectStatementNode) -> bool:
        combined_lineage = SqlExpressionTreeLineage.merge_iterable(
            tuple(x.expr.lineage for x in node.select_columns)
            + ((node.where.lineage,) if node.where else ())
            + tuple(x.expr.lineage for x in node.group_bys)
            + tuple(x.expr.lineage for x in node.order_bys)
        )

        return combined_lineage.contains_string_exprs or combined_lineage.contains_column_alias_exprs

    @staticmethod
    def _select_columns_contain_string_expressions(select_columns: Tuple[SqlSelectColumn, ...]) -> bool:
        combined_lineage = SqlExpressionTreeLineage.merge_iterable(tuple(x.expr.lineage for x in select_columns))

        return len(combined_lineage.string_exprs) > 0

    @staticmethod
    def _select_columns_are_column_references(select_columns: Tuple[SqlSelectColumn, ...]) -> bool:
        for select_column in select_columns:
            if not select_column.expr.as_column_reference_expression:
                return False
        return True

    @staticmethod
    def _select_columns_with_window_functions(select_columns: Tuple[SqlSelectColumn, ...]) -> List[SqlSelectColumn]:
        return [select_column for select_column in select_columns if select_column.expr.as_window_function_expression]

    @staticmethod
    def _select_column_for_alias(column_alias: str, select_columns: Sequence[SqlSelectColumn]) -> SqlSelectColumn:
        for select_column in select_columns:
            if select_column.column_alias == column_alias:
                return select_column
        raise RuntimeError(f"Column alias {repr(column_alias)} not in SELECT columns: {select_columns}")

    @staticmethod
    def _is_simple_source(node: SqlSelectStatementNode) -> bool:
        """Returns true if the node is simple.

        Simple is defined as having no JOINs, WHERE, GROUP BYs, ORDER BYs, LIMIT, AGG functions, and there are no strings in the column
        select. Strings are avoided so that the child node doesn't use the string expression in a group by or cause
        aliasing issues when used in the child query. Aggregate functions are avoided due to the nature of applying on grouped rows which
        is essentially the effect as group bys and should be treated in here as such.

        e.g.

        Reducing

        SELECT
          a.col AS a_col
          , b.col AS b_col
        FROM (
           SELECT
             col AS col -- "col" is a string expression, not a column reference expression.
             , src0.join_key AS join_key
           FROM foo src0
        ) a
        JOIN b
        ON a.join_key = b.join_key

        to

        SELECT
          col AS a_col
          , b.col AS b_col
        FROM foo src0
        JOIN b
        ON a.join_key = b.join_key

        will throw an ambiguous column error on "col".
        """
        for select_column in node.select_columns:
            if select_column.expr.lineage.contains_string_exprs:
                return False
            if select_column.expr.lineage.contains_column_alias_exprs:
                return False
            if select_column.expr.lineage.contains_aggregate_exprs:
                return False
        return (
            len(node.join_descs) == 0
            and len(node.group_bys) == 0
            and len(node.order_bys) == 0
            and not node.limit
            and not node.where
        )

    def _current_node_can_be_reduced(self, node: SqlSelectStatementNode) -> bool:
        """Returns true if the given node can be reduced with the parent node.

        Reducing this node means eliminating the SELECT of this node and merging it with the parent SELECT. This
        checks for the cases where we are able to reduce.
        """
        # If this node has joins, then don't collapse this as it can be complex.
        if len(node.join_descs) > 0:
            return False

        # If the parent node defines CTEs, don't reduce for simplicity. It's possible to improve this by keeping track
        # of the CTE-alias mapping as the SQL plan is traversed and then allowing for reduction if there are no
        # alias collisions (e.g. with other CTEs or the alias in the FROM clause).
        from_clause_node = node.from_source.as_select_node
        if from_clause_node is not None:
            if len(from_clause_node.cte_sources) > 0:
                return False

        # If the parent node is not a SELECT statement, then this can't be collapsed. e.g. with a table as a parent like
        # SELECT foo FROM bar
        from_source_node_as_select_node = node.from_source.as_select_node
        if from_source_node_as_select_node is None:
            return False

        # More conditions where we don't want to collapse. It's not impossible with these cases, but not reducing in
        # these cases for simplicity.

        # Re-writing string expressions / column alias expressions not yet supported, so don't reduce in those cases.
        if SqlRewritingSubQueryReducerVisitor._statement_contains_difficult_expressions(node):
            return False

        # Don't reduce distinct selects
        if from_source_node_as_select_node.distinct:
            return False

        # Skip this case for simplicity of reasoning.
        if len(node.order_bys) > 0 and len(from_source_node_as_select_node.order_bys) > 0:
            return False

        # Skip this case for simplicity of reasoning.
        if len(from_source_node_as_select_node.group_bys) > 0 and len(node.group_bys) > 0:
            return False

        # If there is a column in the parent group by that is not used in the current select statement, don't reduce or it
        # would leave an unselected column in the group by and change the meaning of the query. For example, in the SQL
        # below, reducing would remove the `is_instant` from the select statement.
        #
        # SELECT
        #   bookings
        #   , 2 * bookings AS twice_bookings
        # FROM (
        #   SELECT,
        #     SUM(bookings) AS bookings
        #     , fct_bookings_src.is_instant
        #   FROM demo.fct_bookings fct_bookings_src
        #   GROUP BY fct_bookings_src.is_instant
        # ) src
        #
        # Note: this is not as fine-tuned as it could be. This checks if all parent group bys are used in the current select
        # columns as column reference expressions. If any are used in different types of expressions, we could reduce but
        # won't. This is just limited by the complexity of different expressions that might be used.
        current_select_column_refs = {
            select_column.expr.as_column_reference_expression.col_ref.column_name
            for select_column in node.select_columns
            if select_column.expr.as_column_reference_expression
        }
        all_parent_group_bys_used_in_current_select = True
        for group_by in from_source_node_as_select_node.group_bys:
            parent_group_by_select = SqlGroupByRewritingVisitor._find_matching_select(
                expr=group_by.expr, select_columns=from_source_node_as_select_node.select_columns
            )
            if parent_group_by_select and parent_group_by_select.column_alias not in current_select_column_refs:
                all_parent_group_bys_used_in_current_select = False
        if not all_parent_group_bys_used_in_current_select:
            return False

        # Don't reduce if the ORDER BYs aren't column reference expressions or column alias reference expressions for simplicity.
        # Also for simplicity, the ORDER BY must match one of the SELECT expressions.
        for order_by in node.order_bys or from_source_node_as_select_node.order_bys:
            try:
                if not self._get_matching_column_for_order_by(
                    order_by_expr=order_by.expr, select_columns=node.select_columns
                ):
                    return False
            except RuntimeError:
                return False

        # If the parent has a GROUP BY and this has a WHERE, avoid reducing as the WHERE could reference an
        # aggregation expression.
        if len(from_source_node_as_select_node.group_bys) > 0 and node.where:
            return False

        # If the parent has a GROUP BY, the case where it's easiest to merge this with the parent is if all select
        # columns are column references.
        if len(
            from_source_node_as_select_node.group_bys
        ) > 0 and not SqlRewritingSubQueryReducerVisitor._select_columns_are_column_references(node.select_columns):
            return False

        # If the group by is referencing a window function in the parent node, it can't be reduced.
        # Window functions can't be included in group by.
        parent_column_aliases_with_window_functions = {
            select_column.column_alias
            for select_column in SqlRewritingSubQueryReducerVisitor._select_columns_with_window_functions(
                from_source_node_as_select_node.select_columns
            )
        }
        if len(node.group_bys) > 0 and [
            group_by
            for group_by in node.group_bys
            if (
                group_by.column_alias in parent_column_aliases_with_window_functions
                or (
                    group_by.expr.as_column_reference_expression
                    and group_by.expr.as_column_reference_expression.col_ref.column_name
                    in parent_column_aliases_with_window_functions
                )
                or (
                    group_by.expr.as_string_expression
                    and group_by.expr.as_string_expression.sql_expr in parent_column_aliases_with_window_functions
                )
            )
        ]:
            return False

        # If the parent select node contains string columns, and this has a GROUP BY, don't reduce as string columns
        # can have special meanings in a GROUP BY. For example,
        #
        # SELECT
        #   a.col1 AS col1
        #   , a.col2 AS col2
        # FROM (
        #   SELECT
        #     b.col1 AS col1
        #     , 1 AS col2
        #   FROM foo b
        # ) a
        # GROUP BY a.col2
        #
        # would be rewritten in the current algorithm to
        #
        # SELECT
        #   b.col1 AS col1
        #   , 1 AS col2
        # FROM foo b
        # GROUP by 1
        #
        # However, the GROUP BY would be wrong as the "1" string expression would be interpreted as selecting the 1st
        # item in the SELECT to group by.

        if len(node.group_bys) > 0 and SqlRewritingSubQueryReducerVisitor._select_columns_contain_string_expressions(
            select_columns=from_source_node_as_select_node.select_columns,
        ):
            return False

        return True

    @staticmethod
    def _find_matching_table_alias(node: SqlSelectStatementNode, column_alias: str) -> Optional[str]:
        for select_column in node.select_columns:
            if select_column.column_alias == column_alias:
                column_reference_expr = select_column.expr.as_column_reference_expression
                if column_reference_expr:
                    return column_reference_expr.col_ref.table_alias
        return None

    @staticmethod
    def _get_column_replacements(parent_node: SqlSelectStatementNode, parent_node_alias: str) -> SqlColumnReplacements:
        column_replacements = {}
        for select_column in parent_node.select_columns:
            column_reference = SqlColumnReference(
                table_alias=parent_node_alias,
                column_name=select_column.column_alias,
            )
            column_replacements[column_reference] = select_column.expr

        return SqlColumnReplacements(column_replacements)

    @staticmethod
    def _rewrite_select_columns(
        old_select_columns: Tuple[SqlSelectColumn, ...], column_replacements: SqlColumnReplacements
    ) -> Tuple[SqlSelectColumn, ...]:
        return tuple(
            SqlSelectColumn(expr=x.expr.rewrite(column_replacements), column_alias=x.column_alias)
            for x in old_select_columns
        )

    @staticmethod
    def _rewrite_where(
        column_replacements: SqlColumnReplacements,
        node_where: Optional[SqlExpressionNode] = None,
        parent_node_where: Optional[SqlExpressionNode] = None,
    ) -> Optional[SqlExpressionNode]:
        if node_where is None and parent_node_where is None:
            return None
        elif node_where and parent_node_where is None:
            return node_where.rewrite(column_replacements)
        elif node_where is None and parent_node_where:
            return parent_node_where

        # For type checking. The above conditionals should ensure the below.
        assert node_where
        assert parent_node_where
        return SqlLogicalExpression.create(operator=SqlLogicalOperator.AND, args=(node_where, parent_node_where))

    @staticmethod
    def _find_matching_select_column(
        col_ref: SqlColumnReference, select_columns: Sequence[SqlSelectColumn]
    ) -> Optional[SqlSelectColumn]:
        for select_column in select_columns:
            column_reference_expression = select_column.expr.as_column_reference_expression
            if column_reference_expression and column_reference_expression.col_ref == col_ref:
                return select_column
        return None

    @staticmethod
    def _find_matching_select_column_from_alias_ref_expr(
        col_alias_ref_expr: SqlColumnAliasReferenceExpression, select_columns: Sequence[SqlSelectColumn]
    ) -> Optional[SqlSelectColumn]:
        for select_column in select_columns:
            column_reference_expression = select_column.expr.as_column_reference_expression
            if (
                column_reference_expression
                and column_reference_expression.col_ref.column_name == col_alias_ref_expr.column_alias
            ):
                return select_column
        return None

    def _rewrite_node_with_join(self, node: SqlSelectStatementNode) -> SqlSelectStatementNode:
        """Reduces nodes with joins if the join source is simple to reduce.

        Converts this:

        SELECT
          SUM(bookings_src.bookings) AS bookings
          listings_src.country_latest AS listing__country_latest
          bookings_src.ds AS ds
        FROM (
          SELECT
            fct_bookings_src.booking AS bookings
            , fct_bookings_src.ds AS ds
            , fct_bookings_src.listing_id AS listing
          FROM demo.fct_bookings fct_bookings_src
        ) bookings_src
        JOIN (
          SELECT
            dim_listings_src.country_latest AS country_latest
            , listing_id AS listing
          FROM demo.dim_listings dim_listings_src
        ) listings_src
        ON bookings_src.listing = listings_src.listing
        GROUP BY bookings_src.ds

        to:

        SELECT
          SUM(bookings_src.booking) AS bookings
          dim_listings_src.country_latest AS listing__country_latest
          bookings_src.ds AS ds
        FROM demo.fct_bookings fct_bookings_src
        JOIN demo.dim_listings dim_listings_src
        ON bookings_src.listing_id = dim_listings_src.listing_id
        GROUP BY bookings_src.ds
        """
        # Check that there aren't any duplicates in source aliases, or else there would be a collision when reduced.
        # This check is conservative as it checks for duplicates in this node and parent nodes, but depending on
        # on which sources get reduced, there may not be a collision.
        from_source = node.from_source
        from_source_select = from_source.as_select_node
        from_source_alias = node.from_source_alias

        all_source_aliases = [from_source_alias]
        source_alias_set = {from_source_alias}

        if from_source_select:
            all_source_aliases.append(from_source_select.from_source_alias)
            source_alias_set.add(from_source_select.from_source_alias)

        for join_desc in node.join_descs:
            all_source_aliases.append(join_desc.right_source_alias)
            source_alias_set.add(join_desc.right_source_alias)
            joined_node_select = join_desc.right_source.as_select_node
            if joined_node_select:
                all_source_aliases.append(joined_node_select.from_source_alias)
                source_alias_set.add(joined_node_select.from_source_alias)

        # Some duplicate aliases, so safer to not reduce.
        if len(all_source_aliases) != len(source_alias_set):
            return node

        # See if any of the joined sources can be reduced.
        new_join_descs = []

        clauses_to_rewrite = RewritableSqlClauses(
            select_columns=list(node.select_columns),
            wheres=[node.where] if node.where else [],
            group_bys=list(node.group_bys),
            order_bys=list(node.order_bys),
        )

        # Can't re-write ambiguous expressions.
        if clauses_to_rewrite.contains_ambiguous_exprs:
            return node

        additional_where_clauses = []
        column_replacements_from_all_joins = []

        for join_desc in node.join_descs:
            join_select_node = join_desc.right_source.as_select_node

            # Verifying that it's simple makes it easier to reason about the logic.
            if not join_select_node or not SqlRewritingSubQueryReducerVisitor._is_simple_source(join_select_node):
                new_join_descs.append(join_desc)
                continue

            column_replacements = SqlRewritingSubQueryReducerVisitor._get_column_replacements(
                parent_node=join_select_node,
                parent_node_alias=join_desc.right_source_alias,
            )
            column_replacements_from_all_joins.append(column_replacements)

            new_join_descs.append(
                SqlJoinDescription(
                    right_source=join_select_node.from_source,
                    right_source_alias=join_select_node.from_source_alias,
                    on_condition=(
                        join_desc.on_condition.rewrite(column_replacements) if join_desc.on_condition else None
                    ),
                    join_type=join_desc.join_type,
                )
            )

            if join_select_node.where:
                additional_where_clauses.append(join_select_node.where)

            clauses_to_rewrite.rewrite(column_replacements)

        # While the above re-wrote each individual join when that join was reduced, the ON condition could reference
        # columns from other joins, so the column replacements have to be applied to all join ON conditions.
        for column_replacements in column_replacements_from_all_joins:
            new_join_descs = [
                SqlJoinDescription(
                    right_source=x.right_source,
                    right_source_alias=x.right_source_alias,
                    on_condition=(
                        x.on_condition.rewrite(column_replacements=column_replacements) if x.on_condition else None
                    ),
                    join_type=x.join_type,
                )
                for x in new_join_descs
            ]

        from_source_is_simple = (
            SqlRewritingSubQueryReducerVisitor._is_simple_source(from_source_select) if from_source_select else False
        )
        if from_source_select and from_source_is_simple:
            column_replacements = SqlRewritingSubQueryReducerVisitor._get_column_replacements(
                parent_node=from_source_select,
                parent_node_alias=node.from_source_alias,
            )

            if from_source_select.where:
                additional_where_clauses.append(from_source_select.where)

            clauses_to_rewrite.rewrite(column_replacements=column_replacements)
            # This was already checked in _is_simple_source().
            assert len(from_source_select.join_descs) == 0
            from_source = from_source_select.from_source
            from_source_alias = from_source_select.from_source_alias

            new_join_descs = [
                SqlJoinDescription(
                    right_source=x.right_source,
                    right_source_alias=x.right_source_alias,
                    on_condition=(
                        x.on_condition.rewrite(column_replacements=column_replacements) if x.on_condition else None
                    ),
                    join_type=x.join_type,
                )
                for x in new_join_descs
            ]

        return SqlSelectStatementNode.create(
            description=node.description,
            select_columns=tuple(clauses_to_rewrite.select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            cte_sources=tuple(
                cte_source.with_new_select(cte_source.select_statement.accept(self)) for cte_source in node.cte_sources
            ),
            join_descs=tuple(new_join_descs),
            group_bys=tuple(clauses_to_rewrite.group_bys),
            order_bys=tuple(clauses_to_rewrite.order_bys),
            where=clauses_to_rewrite.combine_wheres(additional_where_clauses),
            limit=node.limit,
            distinct=node.distinct,
        )

    @override
    def visit_cte_node(self, node: SqlCteNode) -> SqlPlanNode:
        return SqlCteNode.create(
            select_statement=node.accept(self),
            cte_alias=node.cte_alias,
        )

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlPlanNode:  # noqa: D102
        node_with_reduced_parents = self._reduce_parents(node)

        if len(node_with_reduced_parents.join_descs) > 0:
            return self._rewrite_node_with_join(node_with_reduced_parents)

        if not self._current_node_can_be_reduced(node_with_reduced_parents):
            return node_with_reduced_parents

        # Note that just because this node can't be reduced doesn't mean the parent node can't. In this example, the
        # outer query can't be reduced because there is a join, but sub-query d can be.
        #
        # SELECT a.bookings, b.listing__country
        # FROM (
        #   SELECT d.bookings, d.listing_id
        #   FROM (
        #     SELECT e.bookings, e.listing_id
        #     FROM fct_bookings
        #   ) d
        #   LIMIT 10
        # ) b
        # JOIN dim_listings c
        # ON a.listing_id = b.listing_id

        from_source_node = node_with_reduced_parents.from_source
        from_source_select_node = from_source_node.as_select_node
        assert (
            from_source_select_node is not None
        ), f"{from_source_select_node=} should be set as `_current_node_can_be_reduced()` returned True"

        # At this point, the query should look similar to
        #
        #     SELECT b.foo, b.baz
        #     FROM (
        #       SELECT 1 AS foo, a.baz FROM bar a
        #       LIMIT 10
        #     ) b
        #     GROUP BY b.foo
        #     ORDER by b.baz
        #     LIMIT 1

        column_replacements = SqlRewritingSubQueryReducerVisitor._get_column_replacements(
            parent_node=from_source_select_node,
            parent_node_alias=node.from_source_alias,
        )
        new_order_bys: List[SqlOrderByDescription] = []
        # Handle ORDER BY differently to avoid this hazard with expression rewriting:
        #
        # SELECT a.bookings AS bookings
        # FROM (
        #   SELECT SUM(b.bookings) AS bookings
        #   FROM fct_bookings b
        # ) a
        # ORDER BY a.bookings
        #
        # ->
        #
        # SELECT SUM(b.bookings) AS bookings
        # FROM fct_bookings b
        # ORDER BY SUM(b.bookings) -- Throws an error in some engines. Instead, use the column alias.

        # We don't reduce if both this node and the parent have an ORDER BY, so here pick the order by from either.
        order_bys_to_use = node_with_reduced_parents.order_bys or from_source_select_node.order_bys
        if order_bys_to_use:
            for order_by_item in order_bys_to_use:
                matching_select_column = self._get_matching_column_for_order_by(
                    order_by_expr=order_by_item.expr, select_columns=node_with_reduced_parents.select_columns
                )
                new_order_bys.append(
                    SqlOrderByDescription(
                        expr=SqlColumnAliasReferenceExpression.create(column_alias=matching_select_column.column_alias),
                        desc=order_by_item.desc,
                    )
                )

        # The limit should be the min of this SELECT limit and the parent SELECT limit.
        new_limit: Optional[int] = node_with_reduced_parents.limit
        if new_limit is None:
            new_limit = from_source_select_node.limit
        elif from_source_select_node.limit is not None:
            new_limit = min(new_limit, from_source_select_node.limit)

        new_group_bys: Tuple[SqlSelectColumn, ...] = ()
        if node.group_bys and from_source_select_node.group_bys:
            raise RuntimeError(
                "Attempting to reduce sub-queries when this and the parent have GROUP BYs. This should have been "
                "prevented by _current_node_can_be_reduced()"
            )
        elif node.group_bys:
            new_group_bys = SqlRewritingSubQueryReducerVisitor._rewrite_select_columns(
                old_select_columns=node.group_bys, column_replacements=column_replacements
            )
        elif from_source_select_node.group_bys:
            new_group_bys = from_source_select_node.group_bys

        return SqlSelectStatementNode.create(
            description="\n".join([from_source_select_node.description, node_with_reduced_parents.description]),
            select_columns=SqlRewritingSubQueryReducerVisitor._rewrite_select_columns(
                old_select_columns=node.select_columns, column_replacements=column_replacements
            ),
            from_source=from_source_select_node.from_source,
            from_source_alias=from_source_select_node.from_source_alias,
            cte_sources=tuple(
                cte_source.with_new_select(cte_source.select_statement.accept(self)) for cte_source in node.cte_sources
            ),
            join_descs=from_source_select_node.join_descs,
            group_bys=new_group_bys,
            order_bys=tuple(new_order_bys),
            where=SqlRewritingSubQueryReducerVisitor._rewrite_where(
                column_replacements=column_replacements,
                node_where=node.where,
                parent_node_where=from_source_select_node.where,
            ),
            limit=new_limit,
            distinct=from_source_select_node.distinct,
        )

    def _get_matching_column_for_order_by(
        self, order_by_expr: SqlExpressionNode, select_columns: Sequence[SqlSelectColumn]
    ) -> SqlSelectColumn:
        order_by_col_ref_expr = order_by_expr.as_column_reference_expression
        order_by_col_alias_ref_expr = order_by_expr.as_column_alias_reference_expression
        if order_by_col_ref_expr:
            matching_select_column = SqlRewritingSubQueryReducerVisitor._find_matching_select_column(
                col_ref=order_by_col_ref_expr.col_ref, select_columns=select_columns
            )
        elif order_by_col_alias_ref_expr:
            matching_select_column = (
                SqlRewritingSubQueryReducerVisitor._find_matching_select_column_from_alias_ref_expr(
                    col_alias_ref_expr=order_by_col_alias_ref_expr, select_columns=select_columns
                )
            )
        else:
            raise RuntimeError(
                f"Expected a column reference expression or a column alias reference expression in ORDER BY but got: {order_by_expr}."
                "This indicates internal misconfiguration."
            )

        if not matching_select_column:
            raise RuntimeError(
                "Did not find matching select column for order by. This indicates internal misconfiguration."
            )
        return matching_select_column

    def visit_table_node(self, node: SqlTableNode) -> SqlPlanNode:  # noqa: D102
        return node

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlPlanNode:  # noqa: D102
        return node

    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> SqlPlanNode:  # noqa: D102
        return SqlCreateTableAsNode.create(
            sql_table=node.sql_table,
            parent_node=node.parent_node.accept(self),
        )


class SqlGroupByRewritingVisitor(SqlPlanNodeVisitor[SqlPlanNode]):
    """Re-writes the GROUP BY to use a SqlColumnAliasReferenceExpression."""

    @staticmethod
    def _find_matching_select(
        expr: SqlExpressionNode, select_columns: Sequence[SqlSelectColumn]
    ) -> Optional[SqlSelectColumn]:
        """Given an expression, find the SELECT column that has the same expression."""
        for select_column in select_columns:
            if select_column.expr.matches(expr):
                return select_column
        return None

    @override
    def visit_cte_node(self, node: SqlCteNode) -> SqlPlanNode:
        return node.with_new_select(node.select_statement.accept(self))

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlPlanNode:  # noqa: D102
        new_group_bys = []
        for group_by in node.group_bys:
            matching_select_column = SqlGroupByRewritingVisitor._find_matching_select(
                group_by.expr, node.select_columns
            )
            if matching_select_column:
                new_group_bys.append(
                    SqlSelectColumn(
                        expr=SqlColumnAliasReferenceExpression.create(column_alias=matching_select_column.column_alias),
                        column_alias=matching_select_column.column_alias,
                    )
                )
            else:
                logger.debug(
                    LazyFormat(
                        lambda: f"Did not find matching select for {group_by} in:\n{indent(node.structure_text())}"
                    )
                )
                new_group_bys.append(group_by)

        return SqlSelectStatementNode.create(
            description=node.description,
            select_columns=node.select_columns,
            from_source=node.from_source.accept(self),
            from_source_alias=node.from_source_alias,
            cte_sources=tuple(
                cte_source.with_new_select(cte_source.select_statement.accept(self)) for cte_source in node.cte_sources
            ),
            join_descs=tuple(
                SqlJoinDescription(
                    right_source=x.right_source.accept(self),
                    right_source_alias=x.right_source_alias,
                    on_condition=x.on_condition,
                    join_type=x.join_type,
                )
                for x in node.join_descs
            ),
            group_bys=tuple(new_group_bys),
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
            distinct=node.distinct,
        )

    def visit_table_node(self, node: SqlTableNode) -> SqlPlanNode:  # noqa: D102
        return node

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlPlanNode:  # noqa: D102
        return node

    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> SqlPlanNode:  # noqa: D102
        return SqlCreateTableAsNode.create(
            sql_table=node.sql_table,
            parent_node=node.parent_node.accept(self),
        )


class SqlRewritingSubQueryReducer(SqlPlanOptimizer):
    """Simplify queries by eliminating sub-queries when possible by rewriting expressions.

     Expressions in the SELECT, GROUP BY, and WHERE are can be rewritten.

    e.g. from

    SELECT b.col0 AS foo
    FROM (
      SELECT SUM(a.col0) AS bar
      FROM table0 a
    ) b
    GROUP BY b.col0

    to

    SELECT SUM(a.col0) AS foo
    FROM table0 a
    GROUP BY foo
    """

    def __init__(self, use_column_alias_in_group_bys: bool = False) -> None:  # noqa: D107
        self._use_column_alias_in_group_bys = use_column_alias_in_group_bys

    def optimize(self, node: SqlPlanNode) -> SqlPlanNode:  # noqa: D102
        result = node.accept(SqlRewritingSubQueryReducerVisitor())
        if self._use_column_alias_in_group_bys:
            return result.accept(SqlGroupByRewritingVisitor())
        return result



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Sequence, Tuple

from typing_extensions import override

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class NoMatchingItemsForMeasure(MetricFlowQueryResolutionIssue):
    """Describes an issue with the query where there are no matching items for a measure.

    This can happen if the user specifies a group-by-item that does not exist or is not available for the measure.
    """

    suggestions: Tuple[str, ...]

    @staticmethod
    def from_parameters(  # noqa: D102
        parent_issues: Sequence[MetricFlowQueryResolutionIssue],
        query_resolution_path: MetricFlowQueryResolutionPath,
        input_suggestions: Sequence[str],
    ) -> NoMatchingItemsForMeasure:
        return NoMatchingItemsForMeasure(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=tuple(parent_issues),
            query_resolution_path=query_resolution_path,
            suggestions=tuple(input_suggestions),
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        lines = [
            f"The given input does not match any of the available group-by-items for\n"
            f"{self.query_resolution_path.last_item.ui_description}. Common issues are:\n",
            indent(
                "* Incorrect names.\n"
                "* No valid join paths exist from the measure to the group-by-item.\n"
                "  (fan-out join support is pending).\n"
                "* There are multiple matching join paths.\n"
                "  (disambiguation support is pending)."
            ),
        ]

        if len(self.suggestions) > 0:
            lines.append(f"\nSuggestions:\n{indent(mf_pformat(list(self.suggestions), max_line_length=80))}")
        return "\n".join(lines)

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> NoMatchingItemsForMeasure:
        return NoMatchingItemsForMeasure(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            suggestions=self.suggestions,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.references import EntityReference, GroupByMetricReference
from typing_extensions import override

from metricflow_semantics.model.semantics.linkable_element import ElementPathKey, LinkableElementType
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.instance_spec import InstanceSpecVisitor, LinkableInstanceSpec
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class GroupByMetricSpec(LinkableInstanceSpec, SerializableDataclass):
    """Metric used in group by or where filter.

    Args:
        element_name: Name of the metric being joined.
        entity_links: Sequence of entities joined to join the metric subquery to the outer query. Last entity is the one
            joining the subquery to the outer query.
        metric_subquery_entity_links: Sequence of entities used in the metric subquery to join the metric to the entity.
    """

    metric_subquery_entity_links: Tuple[EntityReference, ...]

    def __post_init__(self) -> None:
        """The inner query and outer query entity paths must end with the same entity (that's what they join on).

        If no entity links, it's because we're already in the final joined node (no links left).
        """
        assert (
            len(self.metric_subquery_entity_links) > 0
        ), "GroupByMetricSpec must have at least one metric_subquery_entity_link."
        if self.entity_links:
            assert (
                self.metric_subquery_entity_links[-1] == self.entity_links[-1]
            ), "Inner and outer query must have the same last entity link in order to join on that link."

    @property
    def without_first_entity_link(self) -> GroupByMetricSpec:  # noqa: D102
        assert len(self.entity_links) > 0, f"Spec does not have any entity links: {self}"
        return GroupByMetricSpec(
            element_name=self.element_name,
            entity_links=self.entity_links[1:],
            metric_subquery_entity_links=self.metric_subquery_entity_links,
        )

    @property
    def without_entity_links(self) -> GroupByMetricSpec:  # noqa: D102
        return GroupByMetricSpec(
            element_name=self.element_name,
            entity_links=(),
            metric_subquery_entity_links=self.metric_subquery_entity_links,
        )

    @property
    def last_entity_link(self) -> EntityReference:  # noqa: D102
        assert len(self.entity_links) > 0, f"Spec does not have any entity links: {self}"
        return self.entity_links[-1]

    @property
    def metric_subquery_entity_spec(self) -> EntitySpec:
        """Spec for the entity that the metric will be grouped by in the metric subquery."""
        assert (
            len(self.metric_subquery_entity_links) > 0
        ), "GroupByMetricSpec must have at least one metric_subquery_entity_link."
        return EntitySpec(
            element_name=self.metric_subquery_entity_links[-1].element_name,
            entity_links=self.metric_subquery_entity_links[:-1],
        )

    @property
    def qualified_name(self) -> str:
        """Element name prefixed with entity links.

        If same entity links are used in inner & outer query, use standard qualified name (country__bookings).
        Else, specify both sets of entity links (listing__country__user__country__bookings).
        """
        if self.entity_links == self.metric_subquery_entity_links:
            entity_links = self.entity_links
        else:
            entity_links = self.entity_links + self.metric_subquery_entity_links

        return StructuredLinkableSpecName(
            entity_link_names=tuple(entity_link.element_name for entity_link in entity_links),
            element_name=self.element_name,
        ).qualified_name

    def __eq__(self, other: Any) -> bool:  # type: ignore[misc] # noqa: D105
        if not isinstance(other, GroupByMetricSpec):
            return False
        return self.element_name == other.element_name and self.entity_links == other.entity_links

    def __hash__(self) -> int:  # noqa: D105
        return hash((self.element_name, self.entity_links, self.metric_subquery_entity_links))

    @property
    def reference(self) -> GroupByMetricReference:  # noqa: D102
        return GroupByMetricReference(element_name=self.element_name)

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_group_by_metric_spec(self)

    @property
    @override
    def element_path_key(self) -> ElementPathKey:
        return ElementPathKey(
            element_name=self.element_name, element_type=LinkableElementType.METRIC, entity_links=self.entity_links
        )

    def with_entity_prefix(self, entity_prefix: EntityReference) -> GroupByMetricSpec:  # noqa: D102
        return GroupByMetricSpec(
            element_name=self.element_name,
            entity_links=(entity_prefix,) + self.entity_links,
            metric_subquery_entity_links=self.metric_subquery_entity_links,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.references import SemanticModelReference
from metricflow_semantics.assert_one_arg import assert_exactly_one_arg_set
from metricflow_semantics.instances import EntityInstance, InstanceSet, MdoInstance, TimeDimensionInstance
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.specs.column_assoc import ColumnAssociation
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from typing_extensions import override

from metricflow.dataset.dataset_classes import DataSet
from metricflow.sql.sql_plan import (
    SqlPlanNode,
    SqlSelectStatementNode,
)


class SqlDataSet(DataSet):
    """A metric data set along with the associated SQL query node that can be rendered to get those values."""

    def __init__(
        self,
        instance_set: InstanceSet,
        sql_select_node: Optional[SqlSelectStatementNode] = None,
        sql_node: Optional[SqlPlanNode] = None,
    ) -> None:
        """Constructor.

        Args:
            instance_set: Describes the instances in the SQL.
            sql_select_node: The SQL that can be rendered to realize the instance set
        """
        self._sql_select_node = sql_select_node
        self._sql_node = sql_node
        assert_exactly_one_arg_set(sql_select_node=sql_select_node, sql_node=sql_node)
        super().__init__(instance_set=instance_set)

    @property
    def sql_node(self) -> SqlPlanNode:  # noqa: D102
        node_to_return = self._sql_select_node or self._sql_node
        if node_to_return is None:
            raise RuntimeError(
                "This node was not created with a SQL node. This should have been prevented by the initializer."
            )
        return node_to_return

    @property
    def checked_sql_select_node(self) -> SqlSelectStatementNode:
        """If applicable, return a SELECT node that can be used to read data from the given SQL table or SQL query.

        Otherwise, an exception is thrown.
        """
        if self._sql_select_node is None:
            raise RuntimeError(
                str(
                    LazyFormat(
                        f"{self.__class__.__name__} was created with a SQL node that is not a {SqlSelectStatementNode.__name__}",
                        sql_node=self.sql_node.structure_text(),
                    )
                )
            )
        return self._sql_select_node

    def column_associations_for_entity(
        self,
        entity_spec: EntitySpec,
    ) -> Sequence[ColumnAssociation]:
        """Given the name of the entity, return the set of columns associated with it in the data set."""
        matching_instances_with_same_entity_links: List[EntityInstance] = []
        matching_instances_with_different_entity_links: List[EntityInstance] = []
        for linkable_instance in self.instance_set.entity_instances:
            if entity_spec.element_name == linkable_instance.spec.element_name:
                if entity_spec.entity_links == linkable_instance.spec.entity_links:
                    matching_instances_with_same_entity_links.append(linkable_instance)
                else:
                    matching_instances_with_different_entity_links.append(linkable_instance)

        # Prioritize instances with matching entity links, but use mismatched links if matching links not found.
        # Semantic model source data sets might have multiple instances of the same entity, in which case we want the one without
        # links. But group by metric source data sets might only have an instance of the entity with links, and we can join to that.
        matching_instances = matching_instances_with_same_entity_links or matching_instances_with_different_entity_links

        if len(matching_instances) != 1:
            raise RuntimeError(
                f"Expected exactly one matching instance for {entity_spec} in instance set, but found: {matching_instances}. "
                f"All entity instances: {self.instance_set.entity_instances}"
            )
        matching_instance = matching_instances[0]
        if not matching_instance.associated_columns:
            raise RuntimeError(
                f"No associated columns for entity instance {matching_instance} in data set."
                "This indicates internal misconfiguration."
            )

        return matching_instance.associated_columns

    def column_association_for_dimension(
        self,
        dimension_spec: DimensionSpec,
    ) -> ColumnAssociation:
        """Given the name of the dimension, return the set of columns associated with it in the data set."""
        matching_instances = 0
        column_associations_to_return = None
        for dimension_instance in self.instance_set.dimension_instances:
            if dimension_instance.spec == dimension_spec:
                column_associations_to_return = dimension_instance.associated_columns
                matching_instances += 1

        if matching_instances > 1:
            raise RuntimeError(
                f"More than one dimension instance with spec {dimension_spec} in " f"instance set: {self.instance_set}"
            )

        if not column_associations_to_return:
            raise RuntimeError(
                f"No dimension instances with spec {dimension_spec} in instance set: {self.instance_set}"
            )

        return column_associations_to_return[0]

    def instances_for_time_dimensions(
        self, time_dimension_specs: Sequence[TimeDimensionSpec]
    ) -> Tuple[TimeDimensionInstance, ...]:
        """Return the instances associated with these specs in the data set."""
        time_dimension_specs_set = set(time_dimension_specs)
        matching_instances = 0
        instances_to_return: Tuple[TimeDimensionInstance, ...] = ()
        for time_dimension_instance in self.instance_set.time_dimension_instances:
            if time_dimension_instance.spec in time_dimension_specs_set:
                instances_to_return += (time_dimension_instance,)
                matching_instances += 1

        if matching_instances != len(time_dimension_specs_set):
            raise RuntimeError(
                f"Unexpected number of time dimension instances found matching specs.\nSpecs: {time_dimension_specs_set}\n"
                f"Instances: {instances_to_return}"
            )

        return instances_to_return

    def instance_for_time_dimension(self, time_dimension_spec: TimeDimensionSpec) -> TimeDimensionInstance:
        """Given a time dimension spec, return the instance associated with it in the data set."""
        instances = self.instances_for_time_dimensions((time_dimension_spec,))
        if not len(instances) == 1:
            raise RuntimeError(
                f"Unexpected number of time dimension instances found matching specs.\nSpecs: {time_dimension_spec}\n"
                f"Instances: {instances}"
            )
        return instances[0]

    def instance_for_spec(self, spec: InstanceSpec) -> MdoInstance:
        """Given a spec, return the instance associated with it in the data set."""
        instances = self.instance_set.as_tuple
        for instance in instances:
            if instance.spec == spec:
                return instance
        raise RuntimeError(
            str(LazyFormat("Did not find instance matching spec in dataset.", spec=spec, instances=instances))
        )

    def instance_for_column_name(self, column_name: str) -> MdoInstance:
        """Given a spec, return the instance associated with it in the data set."""
        instances = self.instance_set.as_tuple
        for instance in instances:
            if instance.associated_column.column_name == column_name:
                return instance
        raise RuntimeError(
            str(
                LazyFormat(
                    "Did not find instance matching column name in dataset.",
                    column_name=column_name,
                    instances=instances,
                )
            )
        )

    def instance_from_time_dimension_grain_and_date_part(
        self, time_dimension_spec: TimeDimensionSpec
    ) -> TimeDimensionInstance:
        """Find instance in dataset that matches the grain and date part of the given time dimension spec."""
        for time_dimension_instance in self.instance_set.time_dimension_instances:
            if (
                time_dimension_instance.spec.time_granularity == time_dimension_spec.time_granularity
                and time_dimension_instance.spec.date_part == time_dimension_spec.date_part
            ):
                return time_dimension_instance

        raise RuntimeError(
            f"Did not find a time dimension instance with matching grain and date part for spec: {time_dimension_spec}\n"
            f"Instances available: {self.instance_set.time_dimension_instances}"
        )

    def column_association_for_time_dimension(self, time_dimension_spec: TimeDimensionSpec) -> ColumnAssociation:
        """Given the name of the time dimension, return the set of columns associated with it in the data set."""
        return self.instance_for_time_dimension(time_dimension_spec).associated_column

    @property
    @override
    def semantic_model_reference(self) -> Optional[SemanticModelReference]:
        return None

    def annotate(self, alias: str, metric_time_spec: TimeDimensionSpec) -> AnnotatedSqlDataSet:
        """Convert to an AnnotatedSqlDataSet with specified metadata."""
        metric_time_column_name = self.column_association_for_time_dimension(metric_time_spec).column_name
        return AnnotatedSqlDataSet(data_set=self, alias=alias, _metric_time_column_name=metric_time_column_name)


@dataclass(frozen=True)
class AnnotatedSqlDataSet:
    """Class to bind a DataSet to transient properties associated with it at a given point in the SqlQueryPlan."""

    data_set: SqlDataSet
    alias: str
    _metric_time_column_name: Optional[str] = None

    @property
    def metric_time_column_name(self) -> str:
        """Direct accessor for the optional metric time name, only safe to call when we know that value is set."""
        assert (
            self._metric_time_column_name
        ), "Expected a valid metric time dimension name to be associated with this dataset, but did not get one!"
        return self._metric_time_column_name



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Sequence, Sized, Tuple

from typing_extensions import override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.query.group_by_item.path_prefixable import PathPrefixable
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


class MetricFlowQueryIssueType(Enum):
    """Errors prevent the query from running. Later, we'll add warnings."""

    ERROR = "ERROR"


@dataclass(frozen=True)
class MetricFlowQueryIssue:
    """An issue in the query that needs attention from the user."""

    issue_type: MetricFlowQueryIssueType
    # Allows for hierarchical issues that have more detailed context and are more easily generated in a recursive call.
    parent_issues: Tuple[MetricFlowQueryIssue, ...]


@dataclass(frozen=True)
class MetricFlowQueryResolutionIssue(PathPrefixable, ABC):
    """Represents an issue that has come up while resolving user inputs to a query."""

    issue_type: MetricFlowQueryIssueType
    parent_issues: Tuple[MetricFlowQueryResolutionIssue, ...]
    query_resolution_path: MetricFlowQueryResolutionPath

    @abstractmethod
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        """Return a string that describes this issue and is suitable for displaying to the user.

        TBD: Make the associated_input a field of this class.
        """
        raise NotImplementedError


@dataclass(frozen=True)
class MetricFlowQueryResolutionIssueSet(Mergeable, PathPrefixable, Sized):
    """The result of resolving query inputs to specs.

    This implements Sized so that empty issue sets don't have to be printed by mf_pformat().
    """

    issues: Tuple[MetricFlowQueryResolutionIssue, ...] = ()

    @override
    def merge(self, other: MetricFlowQueryResolutionIssueSet) -> MetricFlowQueryResolutionIssueSet:
        return MetricFlowQueryResolutionIssueSet(issues=tuple(self.issues) + tuple(other.issues))

    @override
    @classmethod
    def empty_instance(cls) -> MetricFlowQueryResolutionIssueSet:
        return MetricFlowQueryResolutionIssueSet()

    @property
    def errors(self) -> Sequence[MetricFlowQueryResolutionIssue]:  # noqa: D102
        return tuple(issue for issue in self.issues if issue.issue_type is MetricFlowQueryIssueType.ERROR)

    @property
    def has_errors(self) -> bool:  # noqa: D102
        return len(self.errors) > 0

    def add_issue(self, issue: MetricFlowQueryResolutionIssue) -> MetricFlowQueryResolutionIssueSet:
        """Return a new issue set that is this issue set with the given issue added."""
        return MetricFlowQueryResolutionIssueSet(issues=tuple(self.issues) + (issue,))

    @staticmethod
    def from_issue(issue: MetricFlowQueryResolutionIssue) -> MetricFlowQueryResolutionIssueSet:
        """Return an issue set containing only the given issue."""
        return MetricFlowQueryResolutionIssueSet(issues=(issue,))

    @staticmethod
    def from_issues(  # noqa: D102
        issues: Sequence[MetricFlowQueryResolutionIssue],
    ) -> MetricFlowQueryResolutionIssueSet:  # noqa: D102
        return MetricFlowQueryResolutionIssueSet(issues=tuple(issues))

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> MetricFlowQueryResolutionIssueSet:
        return MetricFlowQueryResolutionIssueSet(
            issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.issues),
        )

    @property
    def has_issues(self) -> bool:  # noqa: D102
        return len(self.issues) > 0

    @override
    def __len__(self) -> int:
        return len(self.issues)



from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.protocols.dimension import Dimension, DimensionType
from dbt_semantic_interfaces.protocols.entity import Entity
from dbt_semantic_interfaces.protocols.measure import Measure
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    EntityReference,
    SemanticModelElementReference,
    SemanticModelReference,
)
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from metricflow_semantics.aggregation_properties import AggregationState
from metricflow_semantics.dag.id_prefix import DynamicIdPrefix, StaticIdPrefix
from metricflow_semantics.dag.sequential_id import SequentialIdGenerator
from metricflow_semantics.instances import (
    DimensionInstance,
    EntityInstance,
    InstanceSet,
    MeasureInstance,
    TimeDimensionInstance,
)
from metricflow_semantics.model.semantics.semantic_model_helper import SemanticModelHelper
from metricflow_semantics.model.spec_converters import MeasureConverter
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.time_dimension_spec import DEFAULT_TIME_GRANULARITY, TimeDimensionSpec
from metricflow_semantics.sql.sql_exprs import (
    SqlColumnReference,
    SqlColumnReferenceExpression,
    SqlDateTruncExpression,
    SqlExpressionNode,
    SqlExtractExpression,
    SqlStringExpression,
)
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.time.granularity import ExpandedTimeGranularity
from metricflow_semantics.time.time_spine_source import TimeSpineSource

from metricflow.dataset.semantic_model_adapter import SemanticModelDataSet
from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.sql.sql_plan import (
    SqlSelectColumn,
    SqlSelectStatementNode,
    SqlTableNode,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DimensionConversionResult:
    """Helper class for returning the results of converting dimensions from a semantic model."""

    dimension_instances: Sequence[DimensionInstance]
    time_dimension_instances: Sequence[TimeDimensionInstance]
    select_columns: Sequence[SqlSelectColumn]


class SemanticModelToDataSetConverter:
    """Converts a semantic model in the model to a data set that can be used with the dataflow plan builder.

    Entity links generally refer to the entities used to join the measure source to the dimension source. For
    example, the dimension name "user_id__device_id__platform" has entity links "user_id" and "device_id" and would
    mean that the measure source was joined by "user_id" to an intermediate source, and then it was joined by
    "device_id" to the source containing the "platform" dimension.
    """

    # Regex for inferring whether an expression for an element is a column reference.
    _SQL_IDENTIFIER_REGEX = re.compile("^[a-zA-Z_][a-zA-Z_0-9]*$")

    def __init__(self, column_association_resolver: ColumnAssociationResolver) -> None:  # noqa: D107
        self._column_association_resolver = column_association_resolver

    def _create_dimension_instance(
        self,
        semantic_model_name: str,
        dimension: Dimension,
        entity_links: Tuple[EntityReference, ...],
    ) -> DimensionInstance:
        """Create a dimension instance from the dimension object in the model."""
        dimension_spec = DimensionSpec(
            element_name=dimension.reference.element_name,
            entity_links=entity_links,
        )
        return DimensionInstance(
            associated_columns=(self._column_association_resolver.resolve_spec(dimension_spec),),
            spec=dimension_spec,
            defined_from=(
                SemanticModelElementReference(
                    semantic_model_name=semantic_model_name,
                    element_name=dimension.reference.element_name,
                ),
            ),
        )

    def _create_time_dimension_instance(
        self,
        element_name: str,
        entity_links: Tuple[EntityReference, ...],
        time_granularity: ExpandedTimeGranularity,
        date_part: Optional[DatePart] = None,
        semantic_model_name: Optional[str] = None,
    ) -> TimeDimensionInstance:
        """Create a time dimension instance from the dimension object from a semantic model in the model."""
        time_dimension_spec = TimeDimensionSpec(
            element_name=element_name,
            entity_links=entity_links,
            time_granularity=time_granularity,
            date_part=date_part,
        )

        return TimeDimensionInstance(
            associated_columns=(self._column_association_resolver.resolve_spec(time_dimension_spec),),
            spec=time_dimension_spec,
            defined_from=(
                (
                    SemanticModelElementReference(
                        semantic_model_name=semantic_model_name,
                        element_name=element_name,
                    ),
                )
                if semantic_model_name
                else ()
            ),
        )

    def _create_entity_instance(
        self,
        semantic_model_name: str,
        entity: Entity,
        entity_links: Tuple[EntityReference, ...],
    ) -> EntityInstance:
        """Create an entity instance from the entity object from a semantic modelin the model."""
        entity_spec = EntitySpec(
            element_name=entity.reference.element_name,
            entity_links=entity_links,
        )
        return EntityInstance(
            associated_columns=(self._column_association_resolver.resolve_spec(entity_spec),),
            spec=entity_spec,
            defined_from=(
                SemanticModelElementReference(
                    semantic_model_name=semantic_model_name,
                    element_name=entity.reference.element_name,
                ),
            ),
        )

    @staticmethod
    def _make_element_sql_expr(
        table_alias: str, element_name: str, element_expr: Optional[str] = None
    ) -> SqlExpressionNode:
        """Create an expression that can be used for reading the element from the semantic model's SQL."""
        if element_expr:
            if SemanticModelToDataSetConverter._SQL_IDENTIFIER_REGEX.match(
                element_expr
            ) and element_expr.upper() not in (
                "TRUE",
                "FALSE",
                "NULL",
            ):
                return SqlColumnReferenceExpression.create(
                    SqlColumnReference(
                        table_alias=table_alias,
                        column_name=element_expr,
                    )
                )
            return SqlStringExpression.create(sql_expr=element_expr)

        return SqlColumnReferenceExpression.create(
            SqlColumnReference(
                table_alias=table_alias,
                column_name=element_name,
            )
        )

    def _convert_measures(
        self,
        semantic_model_name: str,
        measures: Sequence[Measure],
        table_alias: str,
    ) -> Tuple[Sequence[MeasureInstance], Sequence[SqlSelectColumn]]:
        # Convert all elements to instances
        measure_instances = []
        select_columns = []
        for measure in measures or []:
            measure_spec = MeasureConverter.convert_to_measure_spec(measure=measure)
            measure_instance = MeasureInstance(
                associated_columns=(self._column_association_resolver.resolve_spec(measure_spec),),
                spec=measure_spec,
                defined_from=(
                    SemanticModelElementReference(
                        semantic_model_name=semantic_model_name,
                        element_name=measure.reference.element_name,
                    ),
                ),
                aggregation_state=AggregationState.NON_AGGREGATED,
            )
            measure_instances.append(measure_instance)
            select_columns.append(
                SqlSelectColumn(
                    expr=SemanticModelToDataSetConverter._make_element_sql_expr(
                        table_alias=table_alias,
                        element_name=measure.reference.element_name,
                        element_expr=measure.expr,
                    ),
                    column_alias=measure_instance.associated_column.column_name,
                )
            )

        return measure_instances, select_columns

    def _convert_dimensions(
        self,
        semantic_model_name: str,
        dimensions: Sequence[Dimension],
        entity_links: Tuple[EntityReference, ...],
        table_alias: str,
    ) -> DimensionConversionResult:
        dimension_instances = []
        time_dimension_instances = []
        select_columns = []

        for dimension in dimensions or []:
            dimension_select_expr = SemanticModelToDataSetConverter._make_element_sql_expr(
                table_alias=table_alias,
                element_name=dimension.reference.element_name,
                element_expr=dimension.expr,
            )
            if dimension.type == DimensionType.CATEGORICAL:
                dimension_instance = self._create_dimension_instance(
                    semantic_model_name=semantic_model_name,
                    dimension=dimension,
                    entity_links=entity_links,
                )
                dimension_instances.append(dimension_instance)
                select_columns.append(
                    SqlSelectColumn(
                        expr=dimension_select_expr,
                        column_alias=dimension_instance.associated_column.column_name,
                    )
                )
            elif dimension.type == DimensionType.TIME:
                derived_time_dimension_instances, time_select_columns = self._convert_time_dimension(
                    dimension_select_expr=dimension_select_expr,
                    dimension=dimension,
                    semantic_model_name=semantic_model_name,
                    entity_links=entity_links,
                )
                time_dimension_instances += derived_time_dimension_instances
                select_columns += time_select_columns
            else:
                assert False, f"Unhandled dimension type: {dimension.type}"

        return DimensionConversionResult(
            dimension_instances=dimension_instances,
            time_dimension_instances=time_dimension_instances,
            select_columns=select_columns,
        )

    def _convert_time_dimension(
        self,
        dimension_select_expr: SqlExpressionNode,
        dimension: Dimension,
        semantic_model_name: str,
        entity_links: Tuple[EntityReference, ...],
    ) -> Tuple[List[TimeDimensionInstance], List[SqlSelectColumn]]:
        """Converts Dimension objects with type TIME into the relevant DataSet columns.

        Time dimensions require special handling that includes adding additional instances
        and select column statements for each granularity and date part
        """
        time_dimension_instances: List[TimeDimensionInstance] = []
        select_columns: List[SqlSelectColumn] = []

        defined_time_granularity = DEFAULT_TIME_GRANULARITY
        if dimension.type_params and dimension.type_params.time_granularity:
            defined_time_granularity = dimension.type_params.time_granularity

        time_dimension_instance = self._create_time_dimension_instance(
            semantic_model_name=semantic_model_name,
            element_name=dimension.reference.element_name,
            entity_links=entity_links,
            time_granularity=ExpandedTimeGranularity.from_time_granularity(defined_time_granularity),
        )
        time_dimension_instances.append(time_dimension_instance)

        # Until we support minimal granularities, we cannot truncate for
        # any time dimension used as part of a validity window, since a validity window might
        # be stored in seconds while we would truncate to daily.
        if dimension.validity_params:
            select_columns.append(
                SqlSelectColumn(
                    expr=dimension_select_expr,
                    column_alias=time_dimension_instance.associated_column.column_name,
                )
            )
        else:
            select_columns.append(
                self._build_column_for_standard_time_granularity(
                    time_granularity=defined_time_granularity,
                    expr=dimension_select_expr,
                    column_alias=time_dimension_instance.associated_column.column_name,
                )
            )

        new_instances, new_columns = self._build_time_dimension_instances_and_columns(
            semantic_model_name=semantic_model_name,
            defined_time_granularity=defined_time_granularity,
            element_name=dimension.reference.element_name,
            entity_links=entity_links,
            dimension_select_expr=dimension_select_expr,
        )
        time_dimension_instances.extend(new_instances)
        select_columns.extend(new_columns)
        return (time_dimension_instances, select_columns)

    def _build_time_dimension_instances_and_columns(
        self,
        defined_time_granularity: TimeGranularity,
        element_name: str,
        entity_links: Tuple[EntityReference, ...],
        dimension_select_expr: SqlExpressionNode,
        semantic_model_name: Optional[str] = None,
    ) -> Tuple[List[TimeDimensionInstance], List[SqlSelectColumn]]:
        time_dimension_instances: List[TimeDimensionInstance] = []
        select_columns: List[SqlSelectColumn] = []
        # Add time dimensions with a larger granularity for ease in query resolution
        for time_granularity in TimeGranularity:
            if time_granularity.to_int() > defined_time_granularity.to_int():
                time_dimension_instance = self._create_time_dimension_instance(
                    semantic_model_name=semantic_model_name,
                    element_name=element_name,
                    entity_links=entity_links,
                    time_granularity=ExpandedTimeGranularity.from_time_granularity(time_granularity),
                )
                time_dimension_instances.append(time_dimension_instance)

                select_columns.append(
                    self._build_column_for_standard_time_granularity(
                        time_granularity=time_granularity,
                        expr=dimension_select_expr,
                        column_alias=time_dimension_instance.associated_column.column_name,
                    )
                )

        # Add all date part options for easy query resolution
        for date_part in DatePart:
            if date_part.to_int() >= defined_time_granularity.to_int():
                time_dimension_instance = self._create_time_dimension_instance(
                    semantic_model_name=semantic_model_name,
                    element_name=element_name,
                    entity_links=entity_links,
                    time_granularity=ExpandedTimeGranularity.from_time_granularity(defined_time_granularity),
                    date_part=date_part,
                )
                time_dimension_instances.append(time_dimension_instance)

                select_columns.append(
                    SqlSelectColumn(
                        expr=SqlExtractExpression.create(date_part=date_part, arg=dimension_select_expr),
                        column_alias=time_dimension_instance.associated_column.column_name,
                    )
                )

        return (time_dimension_instances, select_columns)

    def _build_column_for_standard_time_granularity(
        self, time_granularity: TimeGranularity, expr: SqlExpressionNode, column_alias: str
    ) -> SqlSelectColumn:
        return SqlSelectColumn(
            expr=SqlDateTruncExpression.create(time_granularity=time_granularity, arg=expr), column_alias=column_alias
        )

    def _create_entity_instances(
        self,
        semantic_model_name: str,
        entities: Sequence[Entity],
        entity_links: Tuple[EntityReference, ...],
        table_alias: str,
    ) -> Tuple[Sequence[EntityInstance], Sequence[SqlSelectColumn]]:
        entity_instances = []
        select_columns = []
        for entity in entities or []:
            # We don't want to create something like user_id__user_id, so skip if the link is the same as the
            # entity.
            if len(entity_links) == 1 and entity.reference == entity_links[0]:
                continue

            entity_instance = self._create_entity_instance(
                semantic_model_name=semantic_model_name,
                entity=entity,
                entity_links=entity_links,
            )

            entity_instances.append(entity_instance)
            select_columns.append(
                SqlSelectColumn(
                    expr=SemanticModelToDataSetConverter._make_element_sql_expr(
                        table_alias=table_alias,
                        element_name=entity.reference.element_name,
                        element_expr=entity.expr,
                    ),
                    column_alias=entity_instance.associated_column.column_name,
                )
            )
        return entity_instances, select_columns

    def create_sql_source_data_set(self, semantic_model: SemanticModel) -> SemanticModelDataSet:
        """Create an SQL source data set from a semantic model in the model."""
        # Gather all instances and columns from all semantic models.
        all_measure_instances: List[MeasureInstance] = []
        all_dimension_instances: List[DimensionInstance] = []
        all_time_dimension_instances: List[TimeDimensionInstance] = []
        all_entity_instances: List[EntityInstance] = []

        all_select_columns: List[SqlSelectColumn] = []
        from_source_alias = SequentialIdGenerator.create_next_id(
            DynamicIdPrefix(prefix=f"{semantic_model.name}_src")
        ).str_value
        # Handle measures
        if len(semantic_model.measures) > 0:
            measure_instances, select_columns = self._convert_measures(
                semantic_model_name=semantic_model.name,
                measures=semantic_model.measures,
                table_alias=from_source_alias,
            )
            all_measure_instances.extend(measure_instances)
            all_select_columns.extend(select_columns)

        # Group by items in the semantic model can be accessed though a subset of the entities defined in the model.
        possible_entity_links: List[Tuple[EntityReference, ...]] = [
            (),
        ]

        for entity_link in SemanticModelHelper.entity_links_for_local_elements(semantic_model):
            possible_entity_links.append((entity_link,))

        # Handle dimensions
        conversion_results = [
            self._convert_dimensions(
                semantic_model_name=semantic_model.name,
                dimensions=semantic_model.dimensions,
                entity_links=entity_links,
                table_alias=from_source_alias,
            )
            for entity_links in possible_entity_links
        ]

        all_dimension_instances.extend(
            [
                dimension_instance
                for conversion_result in conversion_results
                for dimension_instance in conversion_result.dimension_instances
            ]
        )

        all_time_dimension_instances.extend(
            [
                time_dimension_instance
                for conversion_result in conversion_results
                for time_dimension_instance in conversion_result.time_dimension_instances
            ]
        )

        all_select_columns.extend(
            [
                select_column
                for conversion_result in conversion_results
                for select_column in conversion_result.select_columns
            ]
        )

        # Handle entities
        for entity_links in possible_entity_links:
            entity_instances, select_columns = self._create_entity_instances(
                semantic_model_name=semantic_model.name,
                entities=semantic_model.entities,
                entity_links=entity_links,
                table_alias=from_source_alias,
            )
            all_entity_instances.extend(entity_instances)
            all_select_columns.extend(select_columns)

        # Generate the "from" clause depending on whether it's an SQL query or an SQL table.
        from_source = SqlTableNode.create(sql_table=SqlTable.from_string(semantic_model.node_relation.relation_name))

        select_statement_node = SqlSelectStatementNode.create(
            description=f"Read Elements From Semantic Model '{semantic_model.name}'",
            select_columns=tuple(all_select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
        )

        return SemanticModelDataSet(
            semantic_model_reference=SemanticModelReference(semantic_model_name=semantic_model.name),
            instance_set=InstanceSet(
                measure_instances=tuple(all_measure_instances),
                dimension_instances=tuple(all_dimension_instances),
                time_dimension_instances=tuple(all_time_dimension_instances),
                entity_instances=tuple(all_entity_instances),
                metric_instances=(),
            ),
            sql_select_node=select_statement_node,
        )

    def build_time_spine_source_data_set(self, time_spine_source: TimeSpineSource) -> SqlDataSet:
        """Build data set for time spine."""
        from_source_alias = SequentialIdGenerator.create_next_id(StaticIdPrefix.TIME_SPINE_SOURCE).str_value
        base_granularity = time_spine_source.base_granularity
        base_column_name = time_spine_source.base_column

        time_dimension_instances: List[TimeDimensionInstance] = []
        select_columns: List[SqlSelectColumn] = []

        # Build base time dimension instances & columns
        base_time_dimension_instance = self._create_time_dimension_instance(
            element_name=base_column_name,
            entity_links=(),
            time_granularity=ExpandedTimeGranularity.from_time_granularity(base_granularity),
        )
        time_dimension_instances.append(base_time_dimension_instance)
        base_dimension_select_expr = SqlColumnReferenceExpression.from_table_and_column_names(
            table_alias=from_source_alias, column_name=base_column_name
        )
        base_select_column = SqlSelectColumn(
            expr=base_dimension_select_expr, column_alias=base_time_dimension_instance.associated_column.column_name
        )
        select_columns.append(base_select_column)
        new_base_instances, new_base_columns = self._build_time_dimension_instances_and_columns(
            defined_time_granularity=base_granularity,
            element_name=base_column_name,
            entity_links=(),
            dimension_select_expr=base_dimension_select_expr,
        )
        time_dimension_instances.extend(new_base_instances)
        select_columns.extend(new_base_columns)

        # Build custom granularity time dimension instances & columns
        for custom_granularity in time_spine_source.custom_granularities:
            custom_time_dimension_instance = self._create_time_dimension_instance(
                # Use base column name here for ease in building metric_time elements in MetricTimeDimensionTransformNode.
                element_name=base_column_name,
                entity_links=(),
                time_granularity=ExpandedTimeGranularity(
                    name=custom_granularity.name, base_granularity=base_granularity
                ),
            )
            time_dimension_instances.append(custom_time_dimension_instance)
            custom_select_column = SqlSelectColumn(
                expr=SemanticModelToDataSetConverter._make_element_sql_expr(
                    table_alias=from_source_alias, element_name=custom_granularity.parsed_column_name
                ),
                column_alias=custom_time_dimension_instance.associated_column.column_name,
            )
            select_columns.append(custom_select_column)

        return SqlDataSet(
            instance_set=InstanceSet(time_dimension_instances=tuple(time_dimension_instances)),
            sql_select_node=SqlSelectStatementNode.create(
                description=time_spine_source.data_set_description,
                select_columns=tuple(select_columns),
                from_source=SqlTableNode.create(sql_table=time_spine_source.spine_table),
                from_source_alias=from_source_alias,
            ),
        )



from __future__ import annotations

from typing import List, Sequence, Tuple

from metricflow_semantics.model.semantics.linkable_element import LinkableElement
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec


class RenderedSpecTracker:
    """Mutable object that is used to record the specs that were rendered in a where filter.

    This is useful for constructing a WhereFilterSpec as it includes a list of specs required by the filter.
    """

    def __init__(self) -> None:  # noqa: D107
        self._rendered_specs_to_elements: List[Tuple[LinkableInstanceSpec, Sequence[LinkableElement]]] = []

    def record_rendered_spec_to_elements_mapping(
        self, spec_to_elements: Tuple[LinkableInstanceSpec, Sequence[LinkableElement]]
    ) -> None:
        """Records a spec that was rendered in a where filter and can be retrieved later through rendered_specs().

        The mapping to LinkableElements is to facilitate predicate pushdown evaluation on a filter-by-filter basis.
        """
        self._rendered_specs_to_elements.append(spec_to_elements)

    @property
    def rendered_specs_to_elements(self) -> Sequence[Tuple[LinkableInstanceSpec, Sequence[LinkableElement]]]:
        """Returns specs that were recorded by record_rendered_spec()."""
        return self._rendered_specs_to_elements



from __future__ import annotations

import logging
import pprint
from collections.abc import Mapping
from dataclasses import fields, is_dataclass
from enum import Enum
from typing import Any, List, Optional, Sized, Tuple, Union

from dsi_pydantic_shim import BaseModel

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable

logger = logging.getLogger(__name__)


class MetricFlowPrettyFormatter:
    """Creates string representations of objects useful for logging / snapshots."""

    def __init__(
        self,
        indent_prefix: str,
        max_line_length: int,
        include_object_field_names: bool,
        include_none_object_fields: bool,
        include_empty_object_fields: bool,
    ) -> None:
        """See mf_pformat() for argument descriptions."""
        self._indent_prefix = indent_prefix
        if not max_line_length > 0:
            raise ValueError(f"max_line_length must be > 0 as required by pprint.pformat(). Got {max_line_length}")
        self._max_line_width = max_line_length
        self._include_object_field_names = include_object_field_names
        self._include_none_object_fields = include_none_object_fields
        self._include_empty_object_fields = include_empty_object_fields

    @staticmethod
    def _is_pydantic_base_model(obj: Any):  # type:ignore
        # Checking the attribute as the BaseModel check fails for newer version of Pydantic.
        return isinstance(obj, BaseModel) or hasattr(obj, "__config__")

    def _handle_sequence_obj(
        self, list_like_obj: Union[list, tuple, set, frozenset], remaining_line_width: Optional[int]
    ) -> str:
        """Pretty prints a sequence object i.e. list or tuple.

        Args:
            list_like_obj: A list, tuple, set, or frozenset.
            remaining_line_width: If specified, try to make the string representation <= this many columns wide.

        Returns:
            A string representation of the sequence like (1,) or [1, 2].
        """
        if isinstance(list_like_obj, list):
            left_enclose_str = "["
            right_enclose_str = "]"
        elif isinstance(list_like_obj, tuple):
            left_enclose_str = "("
            right_enclose_str = ")"
        elif isinstance(list_like_obj, set) or isinstance(list_like_obj, frozenset):
            left_enclose_str = f"{type(list_like_obj).__name__}("
            right_enclose_str = ")"
            list_like_obj = sorted(self._handle_any_obj(obj, None) for obj in list_like_obj)
        else:
            raise RuntimeError(f"Unhandled type: {type(list_like_obj)}")

        if len(list_like_obj) == 0:
            return f"{left_enclose_str}{right_enclose_str}"

        # See if this object can be printed in one line.
        items_as_str = tuple(self._handle_any_obj(list_item, remaining_line_width=None) for list_item in list_like_obj)
        line_items = [left_enclose_str]
        if len(items_as_str) > 0:
            line_items.extend([", ".join(items_as_str)])
            if len(items_as_str) == 1:
                line_items.append(",")
        line_items.append(right_enclose_str)
        result_without_width_limit = "".join(line_items)

        if remaining_line_width is None or len(result_without_width_limit) <= remaining_line_width:
            return result_without_width_limit

        # The item can't be printed on one line, so do an indented style like:
        """
        [
            1,
            2,
            ...
        ]
        """

        # Convert each item to a pretty string.
        items_as_str = tuple(
            self._handle_any_obj(
                list_item, remaining_line_width=max(0, remaining_line_width - len(self._indent_prefix))
            )
            for list_item in list_like_obj
        )
        lines = [left_enclose_str]

        # item_block is similar to
        """
        1,
        2,
        3,
        """

        item_block = ",\n".join(items_as_str) + ","
        # Indent the item_block
        if len(item_block) > 0:
            lines.append(indent(item_block, indent_prefix=self._indent_prefix))
        lines.append(right_enclose_str)
        return "\n".join(lines)

    def _handle_indented_key_value_item(  # type: ignore[misc]
        self,
        key: Any,
        value: Any,
        key_value_seperator: str,
        is_dataclass_like_object: bool,
        remaining_line_width: Optional[int],
    ) -> str:
        """Convert a key / value for a mapping-like object to a string that should be placed in an indented block.

        Mapping-like objects include dictionaries, dataclasses, and Pydantic models. The output of this method would
        look like:

        "'key': [1, 2, 3]" or "arg=Foo()", etc.

        and the caller of this method would add the actual indent.

        Args:
            key: The object representing the key.
            value: The object representing the value.
            key_value_seperator: The string used to separate the key and the value. e.g. ": " for dicts, "=" for
            dataclasses.
            is_dataclass_like_object: Set this to True if the given value object is a dataclass to handle some printing
            options specific to dataclasses.
            remaining_line_width: If specified, try to make the string representation <= this many columns wide.

        Returns:
            The block that represents the key / value item and goes in between "[" / "]" in the string representation
            of the mapping-like object.
        """
        # See if the string representation can fit on one line. e.g. "'a': [1, 2]"
        if remaining_line_width is None or remaining_line_width > 0:
            result_items_without_limit: List[str] = []
            if is_dataclass_like_object and self._include_object_field_names:
                result_items_without_limit.append(str(key))
            else:
                result_items_without_limit.append(self._handle_any_obj(key, remaining_line_width=None))
            result_items_without_limit.append(key_value_seperator)
            result_items_without_limit.append(self._handle_any_obj(value, remaining_line_width=None))

            result_without_limit = indent("".join(result_items_without_limit), indent_prefix=self._indent_prefix)
            if remaining_line_width is None or len(result_without_limit) <= remaining_line_width:
                return result_without_limit

        # The string representation can't fit on one line - use multiple. e.g.
        """
        'key':
          [1, 2, 3, 4]
        """

        # Create the string for the key.
        result_lines: List[str] = []
        if is_dataclass_like_object and self._include_object_field_names:
            result_lines.append(str(key) + key_value_seperator)
        else:
            # See if the key can be printed on one line. This depends on the length of the value as the key and the
            # the value as at least the first bit of the value is printed on the same line as the key.
            # e.g.
            # "foo"=[
            #   ...
            # ]
            min_length_of_first_value_line = len(self._handle_any_obj(value, remaining_line_width=0).splitlines()[0])

            key_lines = self._handle_any_obj(
                key,
                remaining_line_width=remaining_line_width - len(key_value_seperator) - min_length_of_first_value_line,
            ).splitlines()
            # key_lines would be something like:
            # [
            #     "KeyObject(",
            #     "    a='foo',",
            #     "    b='bar',",
            #     ")",
            # ]

            if len(key_lines) == 1:
                result_lines.append(key_lines[0] + key_value_seperator)
            else:
                # The key needs to be printed in multiple lines. In that case, we want a result where the key value
                # separator is on the last line with the key. e.g.
                #
                # KeyObject(
                #     a='foo',
                #     b='bar',
                # ): ... <value>
                #
                result_lines.extend(key_lines[:-1])
                result_lines.append(key_lines[-1] + key_value_seperator)

        # Combine key and value.

        # Similar to the key, how we print the value depends on whether the value fits on one line or not. e.g.
        # foo=[1, 2, 3]
        #
        # or
        #
        # foo=[
        #   1,
        #   2,
        #   3,
        # ]

        # See if the value fits in the previous line.
        remaining_width_for_value = max(0, remaining_line_width - len(result_lines[-1]))
        value_str = self._handle_any_obj(value, remaining_line_width=remaining_width_for_value)
        value_lines = value_str.splitlines()

        if len(value_lines) <= 1:
            # Value can fit in the previous line
            result_lines[-1] = result_lines[-1] + value_lines[0]
        else:
            # For the multi-line value, we want to print the first line of the value on the same line as the last line
            # of the key.
            result_lines[-1] = result_lines[-1] + value_lines[0]
            result_lines.extend(value_lines[1:])

        return indent("\n".join(result_lines), indent_prefix=self._indent_prefix)

    def _handle_mapping_like_obj(
        self,
        mapping: Mapping,
        left_enclose_str: str,
        key_value_seperator: str,
        right_enclose_str: str,
        is_dataclass_like_object: bool,
        remaining_line_width: Optional[int],
    ) -> str:
        """Convert a mapping-like object to a pretty string.

        This class treats dataclasses as mappings where the field / field values are the keys / values.


        Args:
            mapping: The mapping object to convert.
            left_enclose_str: The string used on the left side to enclose the object. e.g. "{" for dicts or "Foo(" for
            dataclasses.
            key_value_seperator: The string used to separate keys and values. e.g. ": " for dicts, or "=" for
            dataclasses.
            right_enclose_str: The string used on the right side to enclose the object. e.g. "}" for dicts or ")" for
            dataclasses.
            is_dataclass_like_object: Flag to indicate whether this is a dataclass as there are some differences in
            formatting those.
            remaining_line_width: If specified, try to make the string representation <= this many columns wide.

        Returns:
            A string representation of the mapping. e.g. "{'a'=[1, 2]}" or "Foo(a=[1, 2])".
        """
        # Skip key / values depending on the pretty-print configuration.
        if is_dataclass_like_object and not self._include_none_object_fields:
            mapping = {key: value for key, value in mapping.items() if value is not None}

        if is_dataclass_like_object and not self._include_empty_object_fields:
            mapping = {
                key: value
                for key, value in mapping.items()
                if (isinstance(value, Sized) and len(value) > 0) or (not isinstance(value, Sized))
            }

        if len(mapping) == 0:
            return f"{left_enclose_str}{right_enclose_str}"

        # Handle case if the string representation fits on one line.
        if remaining_line_width is None or remaining_line_width > 0:
            comma_separated_items: List[str] = []
            for key, value in mapping.items():
                key_value_str_items: List[str] = []

                if is_dataclass_like_object:
                    if self._include_object_field_names:
                        key_value_str_items.append(str(key))
                        key_value_str_items.append(key_value_seperator)
                else:
                    key_value_str_items.append(self._handle_any_obj(key, remaining_line_width=None))
                    key_value_str_items.append(key_value_seperator)
                key_value_str_items.append(self._handle_any_obj(value, remaining_line_width=None))
                comma_separated_items.append("".join(key_value_str_items))
            result_without_limit = "".join((left_enclose_str, ", ".join(comma_separated_items), right_enclose_str))

            if remaining_line_width is None or len(result_without_limit) <= remaining_line_width:
                return result_without_limit

        # Handle multi-line case.
        mapping_items_as_str = []
        for key, value in mapping.items():
            mapping_items_as_str.append(
                self._handle_indented_key_value_item(
                    key=key,
                    value=value,
                    key_value_seperator=key_value_seperator,
                    is_dataclass_like_object=is_dataclass_like_object,
                    remaining_line_width=(remaining_line_width - len(self._indent_prefix)),
                )
            )
        lines = [left_enclose_str, ",\n".join(mapping_items_as_str) + ",", right_enclose_str]
        return "\n".join(lines)

    def _handle_any_obj(self, obj: Any, remaining_line_width: Optional[int]) -> str:  # type: ignore
        """Convert any object into a pretty string-representation.

        This is called recursively as sequences and mappings have constituent objects of any type.

        Args:
            obj: The object to convert.
            remaining_line_width: If specified, try to make the string representation <= this many columns wide.

        Returns:
            A pretty string-representation of the object.
        """
        if isinstance(obj, Enum):
            return obj.name

        if isinstance(obj, (list, tuple, set, frozenset)):
            return self._handle_sequence_obj(obj, remaining_line_width=remaining_line_width)

        if isinstance(obj, dict):
            return self._handle_mapping_like_obj(
                obj,
                left_enclose_str="{",
                key_value_seperator=": ",
                right_enclose_str="}",
                is_dataclass_like_object=False,
                remaining_line_width=remaining_line_width,
            )

        if isinstance(obj, MetricFlowPrettyFormattable):
            if obj.pretty_format is not None:
                return obj.pretty_format

        if is_dataclass(obj):
            # dataclasses.asdict() seems to exclude None fields, so doing this instead.
            mapping = {field.name: getattr(obj, field.name) for field in fields(obj)}
            return self._handle_mapping_like_obj(
                mapping,
                left_enclose_str=type(obj).__name__ + "(",
                key_value_seperator="=",
                right_enclose_str=")",
                is_dataclass_like_object=True,
                remaining_line_width=remaining_line_width,
            )

        if MetricFlowPrettyFormatter._is_pydantic_base_model(obj):
            mapping = {key: getattr(obj, key) for key in obj.dict().keys()}
            return self._handle_mapping_like_obj(
                mapping,
                left_enclose_str=type(obj).__name__ + "(",
                key_value_seperator="=",
                right_enclose_str=")",
                is_dataclass_like_object=True,
                remaining_line_width=remaining_line_width,
            )

        # Any other object that's not handled.
        return pprint.pformat(obj, width=self._max_line_width, sort_dicts=False)

    def pretty_format(self, obj: Any) -> str:  # type: ignore[misc]
        """Return a pretty string representation of the object that's suitable for logging."""
        return self._handle_any_obj(obj, remaining_line_width=self._max_line_width)


def mf_pformat(  # type: ignore
    obj: Any,
    max_line_length: int = 120,
    indent_prefix: str = "  ",
    include_object_field_names: bool = True,
    include_none_object_fields: bool = False,
    include_empty_object_fields: bool = False,
) -> str:
    """Print objects in a pretty way for logging / test snapshots.

    In Python 3.10, the pretty printer class will support dataclasses, so we can remove this once we're on
    3.10. Also tried the prettyprint package with dataclasses, but that prints full names for the classes
    e.g. a.b.MyClass and it also always added line breaks, even if an object could fit on one line, so
    preferring to not use that for compactness.

    e.g.
        metricflow.specs.DimensionSpec(
            element_name='country',
            entity_links=()
        ),

    Instead, the below will print something like:

        DimensionSpec(element_name='country', entity_links=())

    Also, this simplifies the object representation in some cases (e.g. Enums) and provides options for a more compact
    string. This is an improvement on pformat_big_objects() in dbt-semantic-interfaces to be more compact and easier
    to read.

    Args:
        obj: The object to convert to string.
        max_line_length: If the string representation is going to be longer than this, split into multiple lines.
        indent_prefix: The prefix to use for hierarchical indents.
        include_object_field_names: Include field names when printing objects - e.g. Foo(bar='baz') vs Foo('baz')
        include_none_object_fields: Include fields with a None value - e.g. Foo(bar=None) vs Foo()
        include_empty_object_fields: Include fields that are empty - e.g. Foo(bar=()) vs Foo()

    Returns:
        A string representation of the object that's useful for logging / debugging.
    """
    # Since this is used in logging calls, wrap with except so that a bug here doesn't result in something breaking.
    try:
        formatter = MetricFlowPrettyFormatter(
            indent_prefix=indent_prefix,
            max_line_length=max_line_length,
            include_object_field_names=include_object_field_names,
            include_none_object_fields=include_none_object_fields,
            include_empty_object_fields=include_empty_object_fields,
        )
        return formatter.pretty_format(obj)
    except Exception:
        # This automatically includes the call trace.
        logger.exception("Error pretty printing due to an exception - using str() instead.")
        return str(obj)


def mf_pformat_dict(  # type: ignore
    description: Optional[str] = None,
    obj_dict: Optional[Mapping[str, Any]] = None,
    max_line_length: int = 120,
    indent_prefix: str = "  ",
    include_object_field_names: bool = True,
    include_none_object_fields: bool = False,
    include_empty_object_fields: bool = False,
    preserve_raw_strings: bool = False,
    pad_items_with_newlines: bool = False,
) -> str:
    """Prints many objects in an indented form.

    If `preserve_raw_strings` is set, and a value of the obj_dict is of type str, then use the value itself, not the
    representation of the string. e.g. if value="foo", then "foo" instead of "'foo'". Useful for values that contain
    newlines.

    If `pad_items_with_newlines` is set , each key / value section is padded with newlines.
    """
    description_lines: List[str] = [description] if description is not None else []
    obj_dict = obj_dict or {}
    item_sections = []
    for key, value in obj_dict.items():
        if preserve_raw_strings and isinstance(value, str):
            value_str = value
        else:
            value_str = mf_pformat(
                obj=value,
                max_line_length=max(0, max_line_length - len(indent_prefix)),
                indent_prefix=indent_prefix,
                include_object_field_names=include_object_field_names,
                include_none_object_fields=include_none_object_fields,
                include_empty_object_fields=include_empty_object_fields,
            )

        lines_in_value_str = len(value_str.split("\n"))
        item_section_lines: Tuple[str, ...]
        if lines_in_value_str > 1:
            item_section_lines = (
                f"{key}:",
                indent(
                    value_str,
                    indent_prefix=indent_prefix,
                ),
            )
        else:
            item_section_lines = (f"{key}: {value_str}",)
        item_section = "\n".join(item_section_lines)

        if description is None:
            item_sections.append(item_section)
        else:
            item_sections.append(indent(item_section))

    if pad_items_with_newlines:
        return "\n\n".join(description_lines + item_sections)
    else:
        return "\n".join(description_lines + item_sections)



from __future__ import annotations

from enum import Enum


class SqlJoinType(Enum):
    """Enumerates the different kinds of SQL joins.

    The value is the SQL string to be used when rendering the join.
    """

    LEFT_OUTER = "LEFT OUTER JOIN"
    FULL_OUTER = "FULL OUTER JOIN"
    INNER = "INNER JOIN"
    CROSS_JOIN = "CROSS JOIN"

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}.{self.name}"



from __future__ import annotations

import logging
import time
from collections import defaultdict
from dataclasses import dataclass, field
from functools import cached_property
from typing import Dict, List, Sequence, Set, Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.references import SemanticModelReference
from typing_extensions import override

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.model.semantics.linkable_element import (
    ElementPathKey,
    LinkableDimension,
    LinkableElement,
    LinkableElementType,
    LinkableEntity,
    LinkableMetric,
)
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class LinkableElementSet(SemanticModelDerivation):
    """Container class for storing all linkable elements for a metric.

    TODO: There are similarities with LinkableSpecSet - consider consolidation.
    """

    path_key_to_linkable_dimensions: Dict[ElementPathKey, Tuple[LinkableDimension, ...]] = field(default_factory=dict)
    path_key_to_linkable_entities: Dict[ElementPathKey, Tuple[LinkableEntity, ...]] = field(default_factory=dict)
    path_key_to_linkable_metrics: Dict[ElementPathKey, Tuple[LinkableMetric, ...]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Basic validation for ensuring consistency between path key type and value type."""
        mismatched_dimensions = tuple(
            path_key
            for path_key in self.path_key_to_linkable_dimensions.keys()
            if not path_key.element_type.is_dimension_type
        )
        mismatched_entities = tuple(
            path_key
            for path_key in self.path_key_to_linkable_entities
            if path_key.element_type is not LinkableElementType.ENTITY
        )
        mismatched_metrics = tuple(
            path_key
            for path_key in self.path_key_to_linkable_metrics
            if path_key.element_type is not LinkableElementType.METRIC
        )

        mismatched_elements = {
            "dimensions": mismatched_dimensions,
            "entities": mismatched_entities,
            "metrics": mismatched_metrics,
        }

        assert all(len(mismatches) == 0 for mismatches in mismatched_elements.values()), (
            f"Found one or more elements where the element type defined in the path key does not match the value "
            f"type! Mismatched elements: {mismatched_elements}"
        )

        # There shouldn't be a path key without any concrete items. Can be an issue as specs contained in this set are
        # generated from the path keys.
        for key, value in (
            tuple(self.path_key_to_linkable_dimensions.items())
            + tuple(self.path_key_to_linkable_entities.items())
            + tuple(self.path_key_to_linkable_metrics.items())
        ):
            assert len(value) > 0, f"{key} is empty"

        # There shouldn't be any duplicate specs.
        specs = self.specs
        deduped_specs = set(specs)
        assert len(deduped_specs) == len(specs)
        assert len(deduped_specs) == (
            len(self.path_key_to_linkable_dimensions)
            + len(self.path_key_to_linkable_entities)
            + len(self.path_key_to_linkable_metrics)
        )

        # Check time dimensions have the grain set.
        for path_key, linkable_dimensions in self.path_key_to_linkable_dimensions.items():
            if path_key.element_type is LinkableElementType.TIME_DIMENSION:
                for linkable_dimension in linkable_dimensions:
                    assert (
                        linkable_dimension.time_granularity is not None
                    ), f"{path_key} has a dimension without the time granularity set: {linkable_dimension}"

    @staticmethod
    def merge_by_path_key(linkable_element_sets: Sequence[LinkableElementSet]) -> LinkableElementSet:
        """Combine multiple sets together by the path key.

        If there are elements with the same join key and different element(s) in the tuple of values,
        those elements will be categorized as ambiguous.
        Note this does not deduplicate values, so there may be unambiguous merged sets that appear to have
        multiple values if all one does is a simple length check.
        """
        key_to_linkable_dimensions: Dict[ElementPathKey, List[LinkableDimension]] = defaultdict(list)
        key_to_linkable_entities: Dict[ElementPathKey, List[LinkableEntity]] = defaultdict(list)
        key_to_linkable_metrics: Dict[ElementPathKey, List[LinkableMetric]] = defaultdict(list)

        for linkable_element_set in linkable_element_sets:
            for path_key, linkable_dimensions in linkable_element_set.path_key_to_linkable_dimensions.items():
                key_to_linkable_dimensions[path_key].extend(linkable_dimensions)
            for path_key, linkable_entities in linkable_element_set.path_key_to_linkable_entities.items():
                key_to_linkable_entities[path_key].extend(linkable_entities)
            for path_key, linkable_metrics in linkable_element_set.path_key_to_linkable_metrics.items():
                key_to_linkable_metrics[path_key].extend(linkable_metrics)

        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                path_key: tuple(dimensions) for path_key, dimensions in key_to_linkable_dimensions.items()
            },
            path_key_to_linkable_entities={
                path_key: tuple(entities) for path_key, entities in key_to_linkable_entities.items()
            },
            path_key_to_linkable_metrics={
                path_key: tuple(metrics) for path_key, metrics in key_to_linkable_metrics.items()
            },
        )

    @staticmethod
    def intersection_by_path_key(linkable_element_sets: Sequence[LinkableElementSet]) -> LinkableElementSet:
        """Find the intersection of all elements in the sets by path key.

        This will return the intersection of all path keys defined in the sets, but the union of elements associated
        with each path key. In other words, it filters out path keys (i.e., linkable specs) that are not referenced
        in every set in the input sequence, but it preserves all of the various potentially ambiguous LinkableElement
        instances associated with the path keys that remain.

        This is useful to figure out the common dimensions that are possible to query with multiple metrics. You would
        find the LinkableSpecSet for each metric in the query, then do an intersection of the sets.
        """
        if len(linkable_element_sets) == 0:
            return LinkableElementSet()
        elif len(linkable_element_sets) == 1:
            return linkable_element_sets[0]

        # Find path keys that are common to all LinkableElementSets.
        dimension_path_keys: List[Set[ElementPathKey]] = []
        entity_path_keys: List[Set[ElementPathKey]] = []
        metric_path_keys: List[Set[ElementPathKey]] = []
        for linkable_element_set in linkable_element_sets:
            dimension_path_keys.append(set(linkable_element_set.path_key_to_linkable_dimensions.keys()))
            entity_path_keys.append(set(linkable_element_set.path_key_to_linkable_entities.keys()))
            metric_path_keys.append(set(linkable_element_set.path_key_to_linkable_metrics.keys()))
        common_linkable_dimension_path_keys = set.intersection(*dimension_path_keys) if dimension_path_keys else set()
        common_linkable_entity_path_keys = set.intersection(*entity_path_keys) if entity_path_keys else set()
        common_linkable_metric_path_keys = set.intersection(*metric_path_keys) if metric_path_keys else set()
        # Create a new LinkableElementSet that only includes items where the path key is common to all sets.
        join_path_to_linkable_dimensions: Dict[ElementPathKey, Set[LinkableDimension]] = defaultdict(set)
        join_path_to_linkable_entities: Dict[ElementPathKey, Set[LinkableEntity]] = defaultdict(set)
        join_path_to_linkable_metrics: Dict[ElementPathKey, Set[LinkableMetric]] = defaultdict(set)

        for linkable_element_set in linkable_element_sets:
            for path_key, linkable_dimensions in linkable_element_set.path_key_to_linkable_dimensions.items():
                if path_key in common_linkable_dimension_path_keys:
                    join_path_to_linkable_dimensions[path_key].update(linkable_dimensions)
            for path_key, linkable_entities in linkable_element_set.path_key_to_linkable_entities.items():
                if path_key in common_linkable_entity_path_keys:
                    join_path_to_linkable_entities[path_key].update(linkable_entities)
            for path_key, linkable_metrics in linkable_element_set.path_key_to_linkable_metrics.items():
                if path_key in common_linkable_metric_path_keys:
                    join_path_to_linkable_metrics[path_key].update(linkable_metrics)

        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                path_key: tuple(
                    sorted(
                        dimensions,
                        key=lambda linkable_dimension: (
                            linkable_dimension.semantic_model_origin.semantic_model_name
                            if linkable_dimension.defined_in_semantic_model
                            else ""
                        ),
                    )
                )
                for path_key, dimensions in join_path_to_linkable_dimensions.items()
            },
            path_key_to_linkable_entities={
                path_key: tuple(
                    sorted(
                        entities,
                        key=lambda linkable_entity: linkable_entity.defined_in_semantic_model.semantic_model_name,
                    )
                )
                for path_key, entities in join_path_to_linkable_entities.items()
            },
            path_key_to_linkable_metrics={
                path_key: tuple(sorted(metrics, key=lambda linkable_metric: linkable_metric.element_name))
                for path_key, metrics in join_path_to_linkable_metrics.items()
            },
        )

    def linkable_elements_for_path_key(self, path_key: ElementPathKey) -> Sequence[LinkableElement]:
        """Returns the linkable elements associated with the given path key in this set.

        If the path key does not exist in the set, this silently returns an empty Sequence.
        """
        if path_key in self.path_key_to_linkable_dimensions:
            return self.path_key_to_linkable_dimensions[path_key]
        elif path_key in self.path_key_to_linkable_entities:
            return self.path_key_to_linkable_entities[path_key]
        elif path_key in self.path_key_to_linkable_metrics:
            return self.path_key_to_linkable_metrics[path_key]
        else:
            return tuple()

    def filter(
        self,
        element_filter: LinkableElementFilter,
    ) -> LinkableElementSet:
        """Filter elements in the set.

        First, only elements with at least one property in the "with_any_of" set are retained. Then, any elements with
        a property in "without_any_of" set are removed. Lastly, any elements with all properties in without_all_of
        are removed.
        """
        element_names = element_filter.element_names
        with_any_of = element_filter.with_any_of
        without_any_of = element_filter.without_any_of
        without_all_of = element_filter.without_all_of

        key_to_linkable_dimensions: Dict[ElementPathKey, Tuple[LinkableDimension, ...]] = {}
        key_to_linkable_entities: Dict[ElementPathKey, Tuple[LinkableEntity, ...]] = {}
        key_to_linkable_metrics: Dict[ElementPathKey, Tuple[LinkableMetric, ...]] = {}

        for path_key, linkable_dimensions in self.path_key_to_linkable_dimensions.items():
            if element_names is not None and path_key.element_name not in element_names:
                continue

            filtered_linkable_dimensions = tuple(
                linkable_dimension
                for linkable_dimension in linkable_dimensions
                if len(linkable_dimension.property_set.intersection(with_any_of)) > 0
                and len(linkable_dimension.property_set.intersection(without_any_of)) == 0
                and (
                    len(without_all_of) == 0
                    or linkable_dimension.property_set.intersection(without_all_of) != without_all_of
                )
            )
            if len(filtered_linkable_dimensions) > 0:
                key_to_linkable_dimensions[path_key] = filtered_linkable_dimensions

        for path_key, linkable_entities in self.path_key_to_linkable_entities.items():
            if element_names is not None and path_key.element_name not in element_names:
                continue

            filtered_linkable_entities = tuple(
                linkable_entity
                for linkable_entity in linkable_entities
                if len(linkable_entity.property_set.intersection(with_any_of)) > 0
                and len(linkable_entity.property_set.intersection(without_any_of)) == 0
                and (
                    len(without_all_of) == 0
                    or linkable_entity.property_set.intersection(without_all_of) != without_all_of
                )
            )
            if len(filtered_linkable_entities) > 0:
                key_to_linkable_entities[path_key] = filtered_linkable_entities

        for path_key, linkable_metrics in self.path_key_to_linkable_metrics.items():
            if element_names is not None and path_key.element_name not in element_names:
                continue

            filtered_linkable_metrics = tuple(
                linkable_metric
                for linkable_metric in linkable_metrics
                if len(linkable_metric.property_set.intersection(with_any_of)) > 0
                and len(linkable_metric.property_set.intersection(without_any_of)) == 0
                and (
                    len(without_all_of) == 0
                    or linkable_metric.property_set.intersection(without_all_of) != without_all_of
                )
            )
            if len(filtered_linkable_metrics) > 0:
                key_to_linkable_metrics[path_key] = filtered_linkable_metrics

        return LinkableElementSet(
            path_key_to_linkable_dimensions=key_to_linkable_dimensions,
            path_key_to_linkable_entities=key_to_linkable_entities,
            path_key_to_linkable_metrics=key_to_linkable_metrics,
        )

    @cached_property
    def only_unique_path_keys(self) -> LinkableElementSet:
        """Returns a set that only includes path keys that map to a single distinct element."""
        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                path_key: tuple(set(linkable_dimensions))
                for path_key, linkable_dimensions in self.path_key_to_linkable_dimensions.items()
                if len(set(linkable_dimensions)) <= 1
            },
            path_key_to_linkable_entities={
                path_key: tuple(set(linkable_entities))
                for path_key, linkable_entities in self.path_key_to_linkable_entities.items()
                if len(set(linkable_entities)) <= 1
            },
            path_key_to_linkable_metrics={
                path_key: tuple(set(linkable_metrics))
                for path_key, linkable_metrics in self.path_key_to_linkable_metrics.items()
                if len(set(linkable_metrics)) <= 1
            },
        )

    @cached_property
    @override
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        semantic_model_references: Set[SemanticModelReference] = set()
        for linkable_dimensions in self.path_key_to_linkable_dimensions.values():
            for linkable_dimension in linkable_dimensions:
                semantic_model_references.update(linkable_dimension.derived_from_semantic_models)
        for linkable_entities in self.path_key_to_linkable_entities.values():
            for linkable_entity in linkable_entities:
                semantic_model_references.update(linkable_entity.derived_from_semantic_models)
        for linkable_metrics in self.path_key_to_linkable_metrics.values():
            for linkable_metric in linkable_metrics:
                semantic_model_references.update(linkable_metric.derived_from_semantic_models)

        return sorted(semantic_model_references, key=lambda reference: reference.semantic_model_name)

    @cached_property
    def spec_count(self) -> int:
        """If this is mapped to spec objects, the number of specs that would be produced."""
        return (
            len(self.path_key_to_linkable_dimensions.keys())
            + len(self.path_key_to_linkable_entities.keys())
            + len(self.path_key_to_linkable_metrics.keys())
        )

    @cached_property
    def specs(self) -> Sequence[LinkableInstanceSpec]:
        """Converts the items in a `LinkableElementSet` to their corresponding spec objects."""
        specs: List[LinkableInstanceSpec] = []

        for path_key in (
            tuple(self.path_key_to_linkable_dimensions.keys())
            + tuple(self.path_key_to_linkable_entities.keys())
            + tuple(self.path_key_to_linkable_metrics.keys())
        ):
            specs.append(LinkableElementSet._path_key_to_spec(path_key))

        return specs

    @staticmethod
    def _path_key_to_spec(path_key: ElementPathKey) -> LinkableInstanceSpec:
        """Helper method to convert ElementPathKey instances to LinkableInstanceSpecs.

        This is currently used in the context of switching between ElementPathKeys and LinkableInstanceSpecs
        within a LinkableElementSet, so we leave it here for now.
        """
        if path_key.element_type is LinkableElementType.DIMENSION:
            return DimensionSpec(
                element_name=path_key.element_name,
                entity_links=path_key.entity_links,
            )
        elif path_key.element_type is LinkableElementType.TIME_DIMENSION:
            assert path_key.time_granularity is not None
            return TimeDimensionSpec(
                element_name=path_key.element_name,
                entity_links=path_key.entity_links,
                time_granularity=path_key.time_granularity,
                date_part=path_key.date_part,
            )
        elif path_key.element_type is LinkableElementType.ENTITY:
            return EntitySpec(
                element_name=path_key.element_name,
                entity_links=path_key.entity_links,
            )
        elif path_key.element_type is LinkableElementType.METRIC:
            return GroupByMetricSpec(
                element_name=path_key.element_name,
                entity_links=path_key.entity_links,
                metric_subquery_entity_links=path_key.metric_subquery_entity_links,
            )
        else:
            assert_values_exhausted(path_key.element_type)

    def filter_by_spec_patterns(self, spec_patterns: Sequence[SpecPattern]) -> LinkableElementSet:
        """Filter the elements in the set by the given spec patters.

        Returns a new set consisting of the elements in the `LinkableElementSet` that have a corresponding spec that
        match all the given spec patterns.
        """
        start_time = time.time()

        # Spec patterns need all specs to match properly e.g. `MinimumTimeGrainPattern`.
        matching_specs: Sequence[InstanceSpec] = self.specs

        for spec_pattern in spec_patterns:
            matching_specs = spec_pattern.match(matching_specs)
        specs_to_include = set(matching_specs)

        path_key_to_linkable_dimensions: Dict[ElementPathKey, Tuple[LinkableDimension, ...]] = {}
        path_key_to_linkable_entities: Dict[ElementPathKey, Tuple[LinkableEntity, ...]] = {}
        path_key_to_linkable_metrics: Dict[ElementPathKey, Tuple[LinkableMetric, ...]] = {}

        for path_key, linkable_dimensions in self.path_key_to_linkable_dimensions.items():
            if LinkableElementSet._path_key_to_spec(path_key) in specs_to_include:
                path_key_to_linkable_dimensions[path_key] = linkable_dimensions

        for path_key, linkable_entities in self.path_key_to_linkable_entities.items():
            if LinkableElementSet._path_key_to_spec(path_key) in specs_to_include:
                path_key_to_linkable_entities[path_key] = linkable_entities

        for path_key, linkable_metrics in self.path_key_to_linkable_metrics.items():
            if LinkableElementSet._path_key_to_spec(path_key) in specs_to_include:
                path_key_to_linkable_metrics[path_key] = linkable_metrics

        filtered_elements = LinkableElementSet(
            path_key_to_linkable_dimensions=path_key_to_linkable_dimensions,
            path_key_to_linkable_entities=path_key_to_linkable_entities,
            path_key_to_linkable_metrics=path_key_to_linkable_metrics,
        )
        logger.debug(LazyFormat(lambda: f"Filtering valid linkable elements took: {time.time() - start_time:.2f}s"))
        return filtered_elements

    def filter_by_left_semantic_model(
        self, left_semantic_model_reference: SemanticModelReference
    ) -> LinkableElementSet:
        """Return a `LinkableElementSet` with only elements that have the given left semantic model in the join path."""
        path_key_to_linkable_dimensions: Dict[ElementPathKey, Tuple[LinkableDimension, ...]] = {}
        path_key_to_linkable_entities: Dict[ElementPathKey, Tuple[LinkableEntity, ...]] = {}
        path_key_to_linkable_metrics: Dict[ElementPathKey, Tuple[LinkableMetric, ...]] = {}

        for path_key, linkable_dimensions in self.path_key_to_linkable_dimensions.items():
            path_key_to_linkable_dimensions[path_key] = tuple(
                linkable_dimension
                for linkable_dimension in linkable_dimensions
                if linkable_dimension.join_path.left_semantic_model_reference == left_semantic_model_reference
            )

        for path_key, linkable_entities in self.path_key_to_linkable_entities.items():
            path_key_to_linkable_entities[path_key] = tuple(
                linkable_entity
                for linkable_entity in linkable_entities
                if linkable_entity.join_path.left_semantic_model_reference == left_semantic_model_reference
            )

        for path_key, linkable_metrics in self.path_key_to_linkable_metrics.items():
            path_key_to_linkable_metrics[path_key] = tuple(
                linkable_metric
                for linkable_metric in linkable_metrics
                if linkable_metric.join_path.semantic_model_join_path.left_semantic_model_reference
                == left_semantic_model_reference
            )

        return LinkableElementSet(
            path_key_to_linkable_dimensions=path_key_to_linkable_dimensions,
            path_key_to_linkable_entities=path_key_to_linkable_entities,
            path_key_to_linkable_metrics=path_key_to_linkable_metrics,
        )
